from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.containers import BTreeIndex

SPACE = dm.SpaceApply(
    "cognite_migration", description="Space for the asset-centric to data modeling migration", name="cdf_migration"
)

RESOURCE_VIEW_MAPPING = dm.ContainerApply(
    space=SPACE.space,
    external_id="ResourceViewMapping",
    used_for="node",
    properties={
        "resourceType": dm.ContainerProperty(
            type=dm.Text(max_text_size=255),
            nullable=False,
        ),
        "viewId": dm.ContainerProperty(
            type=dm.data_types.Json(),
            nullable=False,
        ),
        "propertyMapping": dm.ContainerProperty(
            type=dm.data_types.Json(),
            nullable=False,
        ),
    },
)

INSTANCE_SOURCE_CONTAINER = dm.ContainerApply(
    space=SPACE.space,
    external_id="InstanceSource",
    used_for="node",
    properties={
        "resourceType": dm.ContainerProperty(
            type=dm.data_types.Enum(
                values={
                    "timeseries": dm.data_types.EnumValue(),
                    "asset": dm.data_types.EnumValue(),
                    "file": dm.data_types.EnumValue(),
                    "event": dm.data_types.EnumValue(),
                    "sequence": dm.data_types.EnumValue(),
                }
            ),
            nullable=False,
        ),
        "id": dm.ContainerProperty(
            type=dm.data_types.Int64(),
            nullable=False,
        ),
        "dataSetId": dm.ContainerProperty(
            type=dm.data_types.Int64(),
            nullable=True,
        ),
        "classicExternalId": dm.ContainerProperty(
            type=dm.data_types.Text(),
            nullable=True,
        ),
        "preferredConsumerViewId": dm.ContainerProperty(
            type=dm.data_types.Json(),
            nullable=True,
        ),
        "resourceViewMapping": dm.ContainerProperty(
            type=dm.data_types.DirectRelation(container=RESOURCE_VIEW_MAPPING.as_id()), nullable=True
        ),
    },
    indexes={
        "id": BTreeIndex(["id"], cursorable=True),
        "resourceType": BTreeIndex(["resourceType", "id"], cursorable=False),
    },
)
CONTAINERS = [RESOURCE_VIEW_MAPPING, INSTANCE_SOURCE_CONTAINER]

RESOURCE_VIEW_MAPPING_VIEW = dm.ViewApply(
    space=SPACE.space,
    external_id="ResourceViewMapping",
    version="v1",
    name="ResourceViewMapping",
    description="The mapping from asset-centric resources to data modeling view.",
    properties={
        "resourceType": dm.MappedPropertyApply(
            container=RESOURCE_VIEW_MAPPING.as_id(),
            container_property_identifier="resourceType",
        ),
        "viewId": dm.MappedPropertyApply(
            container=RESOURCE_VIEW_MAPPING.as_id(),
            container_property_identifier="viewId",
        ),
        "propertyMapping": dm.MappedPropertyApply(
            container=RESOURCE_VIEW_MAPPING.as_id(),
            container_property_identifier="propertyMapping",
        ),
    },
)

INSTANCE_SOURCE_VIEW = dm.ViewApply(
    space=SPACE.space,
    external_id="InstanceSource",
    version="v1",
    name="InstanceSource",
    description="The source of the instance in asset-centric resources.",
    properties={
        "resourceType": dm.MappedPropertyApply(
            container=INSTANCE_SOURCE_CONTAINER.as_id(),
            container_property_identifier="resourceType",
        ),
        "id": dm.MappedPropertyApply(
            container=INSTANCE_SOURCE_CONTAINER.as_id(),
            container_property_identifier="id",
        ),
        "dataSetId": dm.MappedPropertyApply(
            container=INSTANCE_SOURCE_CONTAINER.as_id(),
            container_property_identifier="dataSetId",
        ),
        "classicExternalId": dm.MappedPropertyApply(
            container=INSTANCE_SOURCE_CONTAINER.as_id(),
            container_property_identifier="classicExternalId",
        ),
        "preferredConsumerViewId": dm.MappedPropertyApply(
            container=INSTANCE_SOURCE_CONTAINER.as_id(),
            container_property_identifier="preferredConsumerViewId",
        ),
        "resourceViewMapping": dm.MappedPropertyApply(
            container=INSTANCE_SOURCE_CONTAINER.as_id(),
            container_property_identifier="resourceViewMapping",
            source=RESOURCE_VIEW_MAPPING_VIEW.as_id(),
        ),
    },
)

INSTANCE_SOURCE_VIEW_ID = INSTANCE_SOURCE_VIEW.as_id()
RESOURCE_VIEW_MAPPING_VIEW_ID = RESOURCE_VIEW_MAPPING_VIEW.as_id()

VIEWS = [RESOURCE_VIEW_MAPPING_VIEW, INSTANCE_SOURCE_VIEW]

COGNITE_MIGRATION_MODEL = dm.DataModelApply(
    space=SPACE.space,
    external_id="CogniteMigration",
    version="v1",
    name="CDF Migration Model",
    description="Data model for migrating asset-centric resources to data modeling resources in CDF.",
    views=[RESOURCE_VIEW_MAPPING_VIEW.as_id(), INSTANCE_SOURCE_VIEW.as_id()],
)

MODEL_ID = COGNITE_MIGRATION_MODEL.as_id()



from ._applications import ChartIO
from ._asset_centric import AssetIO, BaseAssetCentricIO, FileMetadataIO
from ._base import StorageIO, TableStorageIO
from ._data_classes import InstanceIdCSVList, InstanceIdRow, ModelList
from ._instances import InstanceIO
from ._raw import RawIO
from ._selectors import (
    AllChartSelector,
    AssetCentricFileSelector,
    AssetCentricSelector,
    AssetSubtreeSelector,
    ChartFileSelector,
    ChartOwnerSelector,
    ChartSelector,
    DataSetSelector,
    InstanceFileSelector,
    InstanceSelector,
    InstanceViewSelector,
)

__all__ = [
    "AllChartSelector",
    "AssetCentricFileSelector",
    "AssetCentricSelector",
    "AssetIO",
    "AssetSubtreeSelector",
    "BaseAssetCentricIO",
    "ChartFileSelector",
    "ChartIO",
    "ChartOwnerSelector",
    "ChartSelector",
    "DataSetSelector",
    "FileMetadataIO",
    "InstanceFileSelector",
    "InstanceIO",
    "InstanceIdCSVList",
    "InstanceIdRow",
    "InstanceSelector",
    "InstanceViewSelector",
    "ModelList",
    "RawIO",
    "StorageIO",
    "TableStorageIO",
]



from pathlib import Path
from unittest.mock import MagicMock

import pytest

from cognite_toolkit._cdf_tk.exceptions import ToolkitFileNotFoundError
from cognite_toolkit._cdf_tk.utils.file import find_adjacent_files


class TestFindAdjacentFiles:
    @pytest.mark.parametrize(
        "datafile, suffix, available_files, expected_files",
        [
            pytest.param(
                "my_table-part0001.RawRows.ndjson",
                ".Table.yaml",
                ["my_table.Table.yaml", "my_table.Table.json"],
                ["my_table.Table.yaml"],
                id="single matching file",
            ),
            pytest.param(
                "my_table-part0001.RawRows.ndjson",
                ".Table.yaml",
                ["other_file.Table.yaml", "my_table.Table.json"],
                [],
                id="no matching file",
            ),
            pytest.param(
                "my_table-part0001.RawRows.ndjson",
                ".Table.yaml",
                ["my_table.Table.yaml", "my_table.Table.yaml", "my_table.Table.json"],
                ["my_table.Table.yaml", "my_table.Table.yaml"],
                id="multiple matching files",
            ),
            pytest.param(
                "my_table-part0001.RawRows.ndjson",
                ".Table.yaml",
                ["not_my_table.Table.yaml", "my_table2.Table.yaml"],
                [],
                id="similar names but no match",
            ),
            pytest.param(
                "my_table-part0001.RawRows.ndjson",
                ".DoesNotExist",
                ["my_table.Table.yaml", "my_table.Table.json"],
                [],
                id="no matching suffix",
            ),
        ],
    )
    def test_find_adjacent_files(
        self, datafile: str, suffix: str, available_files: list[str], expected_files: list[str]
    ) -> None:
        filepath = MagicMock(spec=Path)
        filepath.is_file.return_value = True
        filepath.name = datafile
        directory = MagicMock(spec=Path)
        filepath.parent = directory
        directory_files: list[Path] = []
        for file in available_files:
            mock_file = MagicMock(spec=Path)
            mock_file.name = file
            directory_files.append(mock_file)
        directory.glob.return_value = directory_files

        result = find_adjacent_files(filepath, suffix=suffix)

        actual = [file.name for file in result]
        assert actual == expected_files

    def test_find_adjacent_files_not_a_file(self):
        filepath = MagicMock(spec=Path)
        filepath.is_file.return_value = False
        filepath.name = "my_table-part0001.RawRows.ndjson"
        with pytest.raises(ToolkitFileNotFoundError):
            find_adjacent_files(filepath, suffix=".Table.yaml")



from functools import partial
from pathlib import Path
from typing import Any

from cognite.client.data_classes._base import T_CogniteResourceList
from rich.console import Console

from cognite_toolkit._cdf_tk.storageio import StorageIO
from cognite_toolkit._cdf_tk.utils.collection import chunker
from cognite_toolkit._cdf_tk.utils.fileio import FileReader
from cognite_toolkit._cdf_tk.utils.producer_worker import ProducerWorkerExecutor
from cognite_toolkit._cdf_tk.utils.useful_types import T_ID, JsonVal, T_Selector, T_WritableCogniteResourceList

from ._base import ToolkitCommand


class UploadCommand(ToolkitCommand):
    """Command for uploading data to CDF from files in a specified directory.

    This command reads files matching a specific pattern in the input directory,
    processes them using a StorageIO instance, and uploads the data to CDF.

    Attributes:
        _MAX_QUEUE_SIZE (int): The maximum size of the queue for processing items.
            Set to 80 to balance memory usage and processing speed.
    """

    _MAX_QUEUE_SIZE = 80

    def upload(
        self,
        io: StorageIO[T_ID, T_Selector, T_CogniteResourceList, T_WritableCogniteResourceList],
        input_dir: Path,
        ensure_configurations: bool,
        dry_run: bool,
        verbose: bool,
    ) -> None:
        """Uploads data from files in the specified input directory to CDF.

        Args:
            io: The StorageIO instance that defines how to upload the data.
            input_dir: The directory containing the files to upload.
            ensure_configurations: If True, creates a configuration for the upload. For example,
                in the case of uploading RAW tables, this will create the RAW database and table.
                For asset-centric, this will create labels and data sets.
            dry_run: If True, performs a dry run without actually uploading the data.
            verbose: If True, prints detailed information about the upload process.

        """
        console = Console()
        cwd = Path.cwd()
        files = list(input_dir.glob(f"*.{io.kind}.*"))
        if verbose:
            input_dir_display = input_dir
            if input_dir.is_relative_to(cwd):
                input_dir_display = input_dir.relative_to(cwd)
            console.print(f"Found {len(files)} files to upload in {input_dir_display.as_posix()!r}.")

        action = "Would upload" if dry_run else "Uploading"
        for file in files:
            file_display = file
            if file_display.is_relative_to(cwd):
                file_display = file_display.relative_to(cwd)
            if verbose:
                console.print(f"{action} {io.display_name} from {file_display.as_posix()!r}")

            selector = io.load_selector(file)
            if ensure_configurations and not dry_run:
                io.ensure_configurations(selector, console)

            reader = FileReader.from_filepath(file)
            executor = ProducerWorkerExecutor[list[dict[str, JsonVal]], T_CogniteResourceList](
                download_iterable=chunker(reader.read_chunks(), io.chunk_size),
                process=io.json_chunk_to_data,
                write=partial(io.upload_items, selector=selector) if not dry_run else self._no_op,
                iteration_count=None,
                max_queue_size=self._MAX_QUEUE_SIZE,
                download_description=f"Reading {file_display.as_posix()!s}",
                process_description="Processing",
                write_description=f"{action} {io.display_name!r}",
                console=console,
            )
            executor.run()
            executor.raise_on_error()
            final_action = "Uploaded" if not dry_run else "Would upload"
            suffix = " successfully" if not dry_run else ""
            console.print(
                f"{final_action} {executor.total_items:,} {io.display_name} from {file_display.as_posix()!r}{suffix}."
            )

    @staticmethod
    def _no_op(_: Any) -> None:
        """A no-operation function used when dry_run is True."""
        pass



from functools import lru_cache

from cognite.client.data_classes.data_modeling import ViewId

from cognite_toolkit._cdf_tk.client.data_classes.migration import AssetCentricToViewMapping, ViewSourceApply

_ASSET_ID = "cdf_asset_mapping"
_EVENT_ID = "cdf_event_mapping"
_TIME_SERIES_ID = "cdf_time_series_mapping"
_FILE_METADATA_ID = "cdf_file_metadata_mapping"


@lru_cache(maxsize=1)
def create_default_mappings() -> list[ViewSourceApply]:
    """Return the default mappings for migration."""
    return [
        ViewSourceApply(
            external_id=_ASSET_ID,
            resource_type="asset",
            view_id=ViewId("cdf_cdm", "CogniteAsset", "v1"),
            mapping=AssetCentricToViewMapping(
                to_property_id={
                    "name": "name",
                    "description": "description",
                    "source": "source",
                    "labels": "tags",
                }
            ),
        ),
        ViewSourceApply(
            external_id=_EVENT_ID,
            resource_type="event",
            view_id=ViewId("cdf_cdm", "CogniteActivity", "v1"),
            mapping=AssetCentricToViewMapping(
                to_property_id={
                    "startTime": "startTime",
                    "endTime": "endTime",
                    "description": "description",
                    "source": "source",
                    "labels": "tags",
                }
            ),
        ),
        ViewSourceApply(
            external_id=_TIME_SERIES_ID,
            resource_type="timeseries",
            view_id=ViewId("cdf_cdm", "CogniteTimeSeries", "v1"),
            mapping=AssetCentricToViewMapping(
                to_property_id={
                    "name": "name",
                    "description": "description",
                    "isStep": "isStep",
                    "isString": "type",
                    "legacyName": "alias",
                    "unit": "sourceUnit",
                    "unitExternalId": "unit",
                    "source": "source",
                }
            ),
        ),
        ViewSourceApply(
            external_id=_FILE_METADATA_ID,
            resource_type="file",
            view_id=ViewId("cdf_cdm", "CogniteFile", "v1"),
            mapping=AssetCentricToViewMapping(
                to_property_id={
                    "name": "name",
                    "description": "description",
                    "source": "source",
                    "labels": "tags",
                    "mimeType": "mimeType",
                    "directory": "directory",
                    "sourceCreatedTime": "sourceCreatedTime",
                    "sourceUpdatedTime": "sourceUpdatedTime",
                }
            ),
        ),
    ]



import gzip
import json
import random
import socket
import sys
import threading
import time
from collections import Counter
from collections.abc import Callable, Hashable, Iterable, MutableMapping
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass, field
from itertools import zip_longest
from queue import Queue
from typing import Generic, Literal, TypeAlias, TypeVar

import requests
import urllib3
from cognite.client import global_config
from cognite.client.utils import _json
from requests.adapters import HTTPAdapter
from requests.structures import CaseInsensitiveDict
from rich.console import Console
from rich.progress import BarColumn, Progress, TaskProgressColumn, TextColumn, TimeRemainingColumn
from urllib3.util.retry import Retry

from cognite_toolkit._cdf_tk.client import ToolkitClientConfig
from cognite_toolkit._cdf_tk.utils.auxiliary import get_current_toolkit_version, get_user_agent

from .collection import chunker
from .useful_types import JsonVal

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


T_ID = TypeVar("T_ID", bound=Hashable)
StatusCode: TypeAlias = int
Count: TypeAlias = int


@dataclass(frozen=True)
class FailedItem(Generic[T_ID]):
    item: T_ID
    status_code: int
    error_message: str


@dataclass(frozen=True)
class SuccessItem(Generic[T_ID]):
    item: T_ID
    status_code: int
    message: str | None = None
    response: dict[str, JsonVal] | None = None  # Optional response data from the server


@dataclass(frozen=True)
class BatchResult(Generic[T_ID]):
    successful_items: list[SuccessItem[T_ID]] = field(default_factory=list)
    failed_items: list[FailedItem[T_ID]] = field(default_factory=list)
    unknown_ids: list[FailedItem[str]] = field(default_factory=list)

    @property
    def total_items(self) -> int:
        return len(self.successful_items) + len(self.failed_items) + len(self.unknown_ids)


@dataclass(frozen=True)
class ProcessorResult(Generic[T_ID]):
    successful_items: list[SuccessItem[T_ID]] = field(default_factory=list)
    failed_items: list[FailedItem[T_ID]] = field(default_factory=list)
    unknown_items: list[FailedItem[str]] = field(default_factory=list)
    error_summary: dict[StatusCode, Count] = field(default_factory=dict)
    producer_error: Exception | None = None

    @property
    def total_successful(self) -> int:
        return len(self.successful_items)

    @property
    def total_failed(self) -> int:
        return len(self.failed_items)

    @property
    def total_processed(self) -> int:
        return self.total_successful + self.total_failed + len(self.unknown_items)

    @property
    def success_rate(self) -> float:
        return len(self.successful_items) / self.total_processed if self.total_processed > 0 else 0.0


@dataclass
class WorkItem:
    items: list[dict]
    connect_attempt: int = 0
    read_attempt: int = 0
    status_attempt: int = 0

    @property
    def total_attempts(self) -> int:
        return self.connect_attempt + self.read_attempt + self.status_attempt


class HTTPProcessor(Generic[T_ID]):
    """A generic HTTP processor for sending items to a specified endpoint in batches.

    This class handles rate limiting, retries, and error handling for HTTP requests.

    Args:
        endpoint_url (str): The URL of the endpoint to send requests to.
        config (ToolkitClientConfig): Configuration for the Toolkit client.
        as_id (Callable[[dict[str, JsonVal]], T_ID]): A function to convert an item to its ID.
        method (Literal["POST", "GET"]): HTTP method to use for requests, default is "POST".
        body_parameters (dict[str, JsonVal] | None): Additional parameters to include in the request body.
        batch_size (int): Number of items per batch, default is 1000.
        max_workers (int): Maximum number of worker threads, default is 8.
        max_retries (int): Maximum number of retries for failed requests, default is 10.
        console (Console | None): Optional console for output, defaults to a new Console instance.

    """

    def __init__(
        self,
        endpoint_url: str,
        config: ToolkitClientConfig,
        as_id: Callable[[dict[str, JsonVal]], T_ID],
        method: Literal["POST", "GET"] = "POST",
        body_parameters: dict[str, JsonVal] | None = None,
        batch_size: int = 1_000,
        max_workers: int = 8,
        max_retries: int = 10,
        console: Console | None = None,
    ):
        self.endpoint_url = endpoint_url
        self.method = method.upper()
        self.body_args = body_parameters
        self.batch_size = batch_size
        self.max_workers = max_workers
        self.max_retries = max_retries
        self.console = console or Console()
        self.as_id = as_id

        self._config = config

        # Thread-safe session for connection pooling
        self.session = self._create_thread_safe_session()

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self, exc_type: type[BaseException] | None, exc_value: BaseException | None, traceback: object | None
    ) -> Literal[False]:
        """Close the session when exiting the context."""
        self.session.close()
        return False  # Do not suppress exceptions

    def _create_thread_safe_session(self) -> requests.Session:
        session = requests.Session()
        adapter = HTTPAdapter(
            pool_connections=self.max_workers,
            pool_maxsize=self.max_workers * 2,
            max_retries=Retry(total=0),  # We handle retries manually
        )
        session.mount("http://", adapter)
        session.mount("https://", adapter)
        return session

    def _create_headers(self) -> MutableMapping[str, str]:
        headers: MutableMapping[str, str] = CaseInsensitiveDict()
        headers.update(requests.utils.default_headers())
        auth_name, auth_value = self._config.credentials.authorization_header()
        headers[auth_name] = auth_value
        headers["content-type"] = "application/json"
        headers["accept"] = "application/json"
        headers["x-cdp-sdk"] = f"CogniteToolkit:{get_current_toolkit_version()}"
        headers["x-cdp-app"] = self._config.client_name
        headers["cdf-version"] = self._config.api_subversion
        if "User-Agent" in headers:
            headers["User-Agent"] += f" {get_user_agent()}"
        else:
            headers["User-Agent"] = get_user_agent()
        if not global_config.disable_gzip:
            headers["Content-Encoding"] = "gzip"
        return headers

    def _prepare_payload(self, work_item: WorkItem) -> str | bytes:
        """
        Prepare the payload for the HTTP request.
        This method should be overridden in subclasses to customize the payload format.
        """
        data: str | bytes
        try:
            data = _json.dumps({"items": work_item.items, **(self.body_args or {})}, allow_nan=False)
        except ValueError as e:
            # A lot of work to give a more human friendly error message when nans and infs are present:
            msg = "Out of range float values are not JSON compliant"
            if msg in str(e):  # exc. might e.g. contain an extra ": nan", depending on build (_json.make_encoder)
                raise ValueError(f"{msg}. Make sure your data does not contain NaN(s) or +/- Inf!").with_traceback(
                    e.__traceback__
                ) from None
            raise

        if not global_config.disable_gzip:
            data = gzip.compress(data.encode())
        return data

    def _worker(
        self,
        work_queue: Queue,
        results_queue: Queue,
    ) -> None:
        while (work_item := work_queue.get()) is not None:
            try:
                response = self._make_request(work_item)
                self._handle_response(response, work_item, work_queue, results_queue)
            except Exception as e:
                self._handle_error(e, work_item, work_queue, results_queue)
            finally:
                work_queue.task_done()
        work_queue.task_done()

    def _make_request(self, work_item: WorkItem) -> requests.Response:
        headers = self._create_headers()
        data = self._prepare_payload(work_item)
        return self.session.request(
            self.method,
            self.endpoint_url,
            data=data,
            headers=headers,
            timeout=self._config.timeout,
            allow_redirects=False,
        )

    def _handle_response(
        self, response: requests.Response, work_item: WorkItem, work_queue: Queue, results_queue: Queue
    ) -> None:
        if 200 <= response.status_code < 300:
            results_queue.put(self._create_success_batch_result(work_item.items, response=response))
        elif response.status_code in {401, 403}:
            self.console.print("[red]Unauthorized request. Please check your credentials.[/red]")
            results_queue.put(self._create_failed_batch_result(work_item.items, response.status_code, response.text))
        elif response.status_code in {400, 409, 422, 502, 503, 504} and len(work_item.items) > 1:
            # 400, 409, 422: There is at least one item that is invalid, split the batch to get all valid items processed
            # 502, 503, 504: Server errors, retry in smaller batches
            mid = len(work_item.items) // 2
            attempts = (
                work_item.status_attempt + 1 if response.status_code in {502, 503, 504} else work_item.status_attempt
            )
            work_queue.put(
                WorkItem(
                    items=work_item.items[:mid],
                    status_attempt=attempts,
                    connect_attempt=work_item.connect_attempt,
                    read_attempt=work_item.read_attempt,
                )
            )
            work_queue.put(
                WorkItem(
                    items=work_item.items[mid:],
                    status_attempt=attempts,
                    connect_attempt=work_item.connect_attempt,
                    read_attempt=work_item.read_attempt,
                )
            )
        elif work_item.status_attempt < self.max_retries and response.status_code in {429, 502, 503, 504}:
            work_item.status_attempt += 1
            time.sleep(self._backoff_time(work_item.total_attempts))
            work_queue.put(work_item)
        else:
            # Permanent failure
            error = response.text
            try:
                body = response.json()
                if "error" in body and "message" in body["error"]:
                    error = body["error"]["message"]
            except ValueError:
                # If the response is not JSON, we keep the original text
                pass
            results_queue.put(self._create_failed_batch_result(work_item.items, response.status_code, error))

    @staticmethod
    def _backoff_time(attempts: int) -> float:
        backoff_time = 0.5 * (2**attempts)
        return min(backoff_time, global_config.max_retry_backoff) * random.uniform(0, 1.0)

    def _handle_error(
        self,
        e: Exception,
        work_item: WorkItem,
        work_queue: Queue,
        results_queue: Queue,
    ) -> None:
        if self._any_exception_in_context_isinstance(
            e, (socket.timeout, urllib3.exceptions.ReadTimeoutError, requests.exceptions.ReadTimeout)
        ):
            error_type = "read"
            work_item.read_attempt += 1
            attempts = work_item.read_attempt
        elif self._any_exception_in_context_isinstance(
            e,
            (
                ConnectionError,
                urllib3.exceptions.ConnectionError,
                urllib3.exceptions.ConnectTimeoutError,
                requests.exceptions.ConnectionError,
            ),
        ):
            error_type = "connect"
            work_item.connect_attempt += 1
            attempts = work_item.connect_attempt
        else:
            error_msg = f"Unexpected exception: {e!s}"
            results_queue.put(self._create_failed_batch_result(work_item.items, 0, error_msg))
            return

        if attempts < self.max_retries:
            time.sleep(self._backoff_time(work_item.total_attempts))
            work_queue.put(work_item)
        else:
            error_msg = f"RequestException after {self.max_retries} {error_type} attempts: {e!s}"
            results_queue.put(self._create_failed_batch_result(work_item.items, 0, error_msg))

    def _create_failed_batch_result(
        self,
        items: list[dict[str, JsonVal]],
        status_code: int,
        error_message: str,
    ) -> BatchResult[T_ID]:
        failed_items: list[FailedItem[T_ID]] = []
        unknown_items: list[FailedItem[str]] = []
        for item in items:
            try:
                item_id = self.as_id(item)
            except Exception as e:
                unknown_items.append(self._create_unknown_item(item, status_code, error_message, e))
            else:
                failed_items.append(FailedItem(item_id, status_code, error_message))
        return BatchResult(failed_items=failed_items, unknown_ids=unknown_items)

    def _create_success_batch_result(
        self, items: list[dict[str, JsonVal]], response: requests.Response, message: str | None = None
    ) -> BatchResult[T_ID]:
        status_code = response.status_code
        response_items = self._parse_response_items(response)
        has_printed_warning = False
        success_items: list[SuccessItem[T_ID]] = []
        unknown_items: list[FailedItem[str]] = []
        failed_items: list[FailedItem[T_ID]] = []
        for item, response_item in zip_longest(items, response_items, fillvalue=None):
            if item is None:
                if not has_printed_warning:
                    self.console.print("[red]Got more response 'items' than request items.[/red]")
                    has_printed_warning = True
                continue

            try:
                item_id = self.as_id(item)
            except Exception as e:
                unknown_items.append(self._create_unknown_item(item, status_code, message, e))
            else:
                if response_item is None:
                    failed_items.append(
                        FailedItem(item=item_id, status_code=status_code, error_message="Response item is None")
                    )
                else:
                    success_items.append(
                        SuccessItem(item=item_id, response=response_item, status_code=status_code, message=message)
                    )
        return BatchResult(successful_items=success_items, unknown_ids=unknown_items, failed_items=failed_items)

    def _parse_response_items(self, response: requests.Response) -> list[dict[str, JsonVal]]:
        try:
            response_items = response.json().get("items", [])
        except (json.JSONDecodeError, TypeError) as e:
            self.console.print(f"[red]Failed to decode JSON response: {e!s}[/red]")
            response_items = []
        if not isinstance(response_items, list):
            self.console.print("[red]Response 'items' is not a list. Skipping response.[/red]")
            response_items = []
        return response_items

    @staticmethod
    def _create_unknown_item(
        item: dict[str, JsonVal], status_code: int, error_message: str | None, exception: Exception
    ) -> FailedItem[str]:
        try:
            item_repr = str(item)[:50]
        except Exception:
            item_repr = "<unrepresentable item>"
        return FailedItem(
            f"as_id failed for item {item_repr} error {exception!s}.", status_code, error_message or "as_id failed"
        )

    @staticmethod
    def _aggregate_results(results: list[BatchResult[T_ID]], producer_error: Exception | None) -> ProcessorResult[T_ID]:
        successful_items = [item for r in results for item in r.successful_items]
        failed_items = [item for r in results for item in r.failed_items]
        unknown_ids = [item for r in results for item in r.unknown_ids]
        error_summary: dict[int, int] = Counter()
        for item in failed_items:
            error_summary[item.status_code] += 1
        for unknown_item in unknown_ids:
            error_summary[unknown_item.status_code] += 1

        return ProcessorResult(
            unknown_items=unknown_ids,
            successful_items=successful_items,
            failed_items=failed_items,
            error_summary=error_summary,
            producer_error=producer_error,
        )

    @classmethod
    def _any_exception_in_context_isinstance(
        cls, exc: BaseException, exc_types: tuple[type[BaseException], ...] | type[BaseException]
    ) -> bool:
        """requests does not use the "raise ... from ..." syntax, so we need to access the underlying exceptions using
        the __context__ attribute.
        """
        if isinstance(exc, exc_types):
            return True
        if exc.__context__ is None:
            return False
        return cls._any_exception_in_context_isinstance(exc.__context__, exc_types)


class HTTPIterableProcessor(HTTPProcessor[T_ID]):
    """A generic HTTP batch processor for sending items to a specified endpoint from an iterable source.

    This class handles batching, rate limiting, retries, and error handling for HTTP requests.

    Args:
        endpoint_url (str): The URL of the endpoint to send requests to.
        config (ToolkitClientConfig): Configuration for the Toolkit client.
        as_id (Callable[[dict[str, JsonVal]], T_ID]): A function to convert an item to its ID.
        method (Literal["POST", "GET"]): HTTP method to use for requests, default is "POST".
        body_parameters (dict[str, JsonVal] | None): Additional parameters to include in the request body.
        batch_size (int): Number of items per batch, default is 1000.
        max_workers (int): Maximum number of worker threads, default is 8.
        max_retries (int): Maximum number of retries for failed requests, default is 10.
        console (Console | None): Optional console for output, defaults to a new Console instance.
        description (str): Description for the progress bar, default is "Processing items".

    """

    def __init__(
        self,
        endpoint_url: str,
        config: ToolkitClientConfig,
        as_id: Callable[[dict[str, JsonVal]], T_ID],
        method: Literal["POST", "GET"] = "POST",
        body_parameters: dict[str, JsonVal] | None = None,
        batch_size: int = 1_000,
        max_workers: int = 8,
        max_retries: int = 10,
        console: Console | None = None,
        description: str = "Processing items",
    ):
        super().__init__(
            endpoint_url=endpoint_url,
            config=config,
            as_id=as_id,
            method=method,
            body_parameters=body_parameters,
            batch_size=batch_size,
            max_workers=max_workers,
            max_retries=max_retries,
            console=console,
        )
        self._produced_count = 0
        self._process_exception: Exception | None = None
        self.description = description

    def _producer(self, items_iterator: Iterable[dict[str, JsonVal]], work_queue: Queue) -> None:
        batch: list[dict] = []
        try:
            for item in items_iterator:
                batch.append(item)
                if len(batch) >= self.batch_size:
                    while work_queue.qsize() >= self.max_workers * 2:
                        # Wait for space in the queue
                        time.sleep(0.1)
                    work_queue.put(WorkItem(items=batch))
                    self._produced_count += len(batch)
                    batch = []
            if batch:
                work_queue.put(WorkItem(items=batch))
                self._produced_count += len(batch)
        except Exception as e:
            self.console.print(f"[red]Error in producer thread: {e!s}[/red]")
            # If an error occurs, we still want to put the items in the queue to avoid losing them
            if batch:
                work_queue.put(WorkItem(items=batch))
                self._produced_count += len(batch)
            self._process_exception = e

    def process(self, items: Iterable[dict[str, JsonVal]], total_items: int | None = None) -> ProcessorResult[T_ID]:
        """Process items in batches using multiple worker threads.

        Args:
            items (Iterable[dict[str, JsonVal]]): An iterable of items to process.
            total_items (int | None): Total number of items to process, used for progress tracking. The tracking
                will only show number of items processed, not the number of items in the queue.

        Returns:
            ProcessorResult[T_ID]: The result of processing the items, including successful and failed items.

        """
        work_queue: Queue[WorkItem | None] = Queue()
        results_queue: Queue[BatchResult[T_ID]] = Queue()

        self._produced_count = 0
        self._process_exception = None
        producer_thread = threading.Thread(target=self._producer, args=(items, work_queue), daemon=True)
        producer_thread.start()
        batch_results: list[BatchResult[T_ID]] = []
        with Progress(
            TextColumn("[bold blue]{task.description}"),
            BarColumn(),
            TaskProgressColumn(),
            TextColumn("{task.fields[processed_items]} items processed"),
            TimeRemainingColumn(),
            console=self.console,
        ) as progress:
            task = progress.add_task(self.description, total=total_items, processed_items=0)
            with ThreadPoolExecutor(max_workers=self.max_workers, thread_name_prefix="batch_worker") as executor:
                # Start workers
                for _ in range(self.max_workers):
                    executor.submit(self._worker, work_queue, results_queue)

                producer_thread.join()
                if self._produced_count == 0:
                    self.console.print("[yellow]No items to process.[/yellow]")
                else:
                    work_queue.join()
                    processed_count = 0
                    while processed_count < self._produced_count:
                        result = results_queue.get()
                        batch_results.append(result)
                        processed_count += result.total_items
                        progress.update(task, processed_items=processed_count)
                for _ in range(self.max_workers):
                    work_queue.put(None)
        return self._aggregate_results(batch_results, self._process_exception)


class HTTPBatchProcessor(HTTPProcessor[T_ID]):
    """An HTTP processor for processing items in batches.

    This class handles batching, rate limiting, retries, and error handling for HTTP requests.

    Args:
        endpoint_url (str): The URL of the endpoint to send requests to.
        config (ToolkitClientConfig): Configuration for the Toolkit client.
        as_id (Callable[[dict[str, JsonVal]], T_ID]): A function to convert an item to its ID.
        result_processor (Callable[[BatchResult[T_ID]], None]): A callable that processes the result of each batch.
        method (Literal["POST", "GET"]): HTTP method to use for requests, default is "POST".
        body_parameters (dict[str, JsonVal] | None): Additional parameters to include in the request body.
        batch_size (int): Number of items per batch, default is 1000.
        max_workers (int): Maximum number of worker threads, default is 8.
        max_retries (int): Maximum number of retries for failed requests, default is 10.
        console (Console | None): Optional console for output, defaults to a new Console instance.

    """

    def __init__(
        self,
        endpoint_url: str,
        config: ToolkitClientConfig,
        as_id: Callable[[dict[str, JsonVal]], T_ID],
        result_processor: Callable[[BatchResult[T_ID]], None],
        method: Literal["POST", "GET"] = "POST",
        body_parameters: dict[str, JsonVal] | None = None,
        batch_size: int = 1_000,
        max_workers: int = 8,
        max_retries: int = 10,
        console: Console | None = None,
    ):
        super().__init__(
            endpoint_url=endpoint_url,
            config=config,
            as_id=as_id,
            method=method,
            body_parameters=body_parameters,
            batch_size=batch_size,
            max_workers=max_workers,
            max_retries=max_retries,
            console=console,
        )
        self.result_processor = result_processor
        self._work_queue: Queue[WorkItem | None] | None = None
        self._result_queue: Queue[BatchResult[T_ID] | None] | None = None
        self._worker_threads: list[threading.Thread] = []
        self._result_thread: threading.Thread | None = None

    def __enter__(self) -> Self:
        """Enter the context manager, initializing the work and result queues."""
        try:
            # Limiting the queue size to avoid excessive memory usage
            self._work_queue = Queue(self.max_workers * 2)
            self._result_queue = Queue()
            self._worker_threads = [
                threading.Thread(target=self._worker, args=(self._work_queue, self._result_queue), daemon=True)
                for _ in range(self.max_workers)
            ]
            self._result_thread = threading.Thread(target=self._result_processor, daemon=True)
            self._result_thread.start()
            for thread in self._worker_threads:
                thread.start()
        except (RuntimeError, OSError) as e:
            self.console.print(f"[red]Error initializing processor: {e!s}[/red]")
            self._stop()
            raise RuntimeError("Failed to initialize the processor.") from e
        return self

    def __exit__(
        self, exc_type: type[BaseException] | None, exc_value: BaseException | None, traceback: object | None
    ) -> Literal[False]:
        """Exit the context manager, stopping the worker threads and closing the queues."""
        self._stop()
        return False

    def add_items(self, items: Iterable[dict[str, JsonVal]]) -> None:
        """Add items to the processor for processing.

        Args:
            items (Iterable[dict[str, JsonVal]]): An iterable of items to process.
        """
        if self._work_queue is None:
            raise RuntimeError("Processor is not initialized. Please use the context manager to initialize it.")
        for chunk in chunker(items, self.batch_size):
            self._work_queue.put(WorkItem(items=chunk))

    def _stop(self) -> None:
        """Stop the processor, joining all worker threads and closing the queues."""
        if self._work_queue is not None:
            for _ in self._worker_threads:
                self._work_queue.put(None)
            for thread in self._worker_threads:
                thread.join()
            self._work_queue = None
        if self._result_queue is not None:
            self._result_queue.put(None)
            if self._result_thread and self._result_thread.is_alive():
                self._result_thread.join()
            self._result_queue = None

    def _result_processor(self) -> None:
        """Process results from the result queue and aggregate them."""
        if self._result_queue is None:
            raise RuntimeError("Result queue is not initialized. Please use the context manager to initialize it.")
        while (result := self._result_queue.get()) is not None:
            try:
                self.result_processor(result)
            except Exception as e:
                # The result processor is user-defined, so we need to catch any exceptions
                self.console.print(f"[red]Error processing result: {e!s}[/red]")
                self.console.print_exception()
            finally:
                self._result_queue.task_done()
        self._result_queue.task_done()



import sys
from collections.abc import Collection, Iterator, Sequence
from pathlib import Path
from typing import Any, SupportsIndex, overload

from cognite.client.data_classes.data_modeling import NodeId, ViewId
from cognite.client.utils._text import to_camel_case
from pydantic import BaseModel, field_validator
from pydantic_core import ValidationError
from rich.console import Console

from cognite_toolkit._cdf_tk.client.data_classes.migration import AssetCentricId
from cognite_toolkit._cdf_tk.client.data_classes.pending_instances_ids import PendingInstanceId
from cognite_toolkit._cdf_tk.commands._migrate.default_mappings import create_default_mappings
from cognite_toolkit._cdf_tk.exceptions import (
    ToolkitValueError,
)
from cognite_toolkit._cdf_tk.tk_warnings import LowSeverityWarning
from cognite_toolkit._cdf_tk.utils.collection import humanize_collection
from cognite_toolkit._cdf_tk.utils.fileio import CSVReader, SchemaColumn

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class MigrationMapping(BaseModel, alias_generator=to_camel_case, extra="ignore"):
    """The mapping between an asset-centric ID and a data modeling instance ID.
    Args
        resource_type (str): The asset-centric type of the resource (e.g., "asset", "event", "timeseries").
        instance_id (NodeId): The target NodeId in data modeling.
        id (int): The asset-centric ID of the resource.
        data_set_id (int | None): The data set ID of the resource. This is used to validate access to the resource.
        ingestion_view (str | None): The ingestion view name. This is the view mapping that will be used to
            ingest the resource into data modeling.
        preferred_consumer_view (ViewId | None): The preferred consumer view for the resource. This is used in
           for example, the Canvas migration to determine which view to use for the resource.
    """

    resource_type: str
    instance_id: NodeId
    id: int
    data_set_id: int | None = None
    ingestion_view: str | None = None
    preferred_consumer_view: ViewId | None = None

    def get_ingestion_view(self) -> str:
        """Get the ingestion view for the mapping. If not specified, return the default ingestion view."""
        if self.ingestion_view:
            return self.ingestion_view

        default_mappings = create_default_mappings()
        for mapping in default_mappings:
            if mapping.resource_type == self.resource_type:
                return mapping.external_id
        raise ToolkitValueError(f"No default ingestion view specified for resource type '{self.resource_type}'")

    def as_asset_centric_id(self) -> AssetCentricId:
        return AssetCentricId(resource_type=self.resource_type, id_=self.id)  # type: ignore[arg-type]

    @field_validator("data_set_id", "ingestion_view", mode="before")
    def _empty_string_to_none(cls, v: Any) -> Any:
        if isinstance(v, str) and not v.strip():
            return None
        return v

    @field_validator("preferred_consumer_view", mode="before")
    def _validate_preferred_consumer_view(cls, v: Any) -> Any:
        if isinstance(v, dict):
            return ViewId.load(v)
        return v

    @field_validator("instance_id", mode="before")
    def _validate_instance_id(cls, v: Any) -> Any:
        if isinstance(v, dict):
            return NodeId.load(v)
        return v


class MigrationMappingList(list, Sequence[MigrationMapping]):
    REQUIRED_HEADER = (
        SchemaColumn("id", "integer"),
        SchemaColumn("space", "string"),
        SchemaColumn("externalId", "string"),
    )
    OPTIONAL_HEADER = (
        SchemaColumn("dataSetId", "integer"),
        SchemaColumn("ingestionView", "string"),
        SchemaColumn("consumerViewSpace", "string"),
        SchemaColumn("consumerViewExternalId", "string"),
        SchemaColumn("consumerViewVersion", "string"),
    )

    def __init__(
        self,
        collection: Collection[MigrationMapping] | None = None,
        error_by_row_no: dict[int, ValidationError] | None = None,
    ) -> None:
        super().__init__(collection or [])
        self.error_by_row_no = error_by_row_no or {}

    def __iter__(self) -> Iterator[MigrationMapping]:
        return super().__iter__()

    @overload
    def __getitem__(self, index: SupportsIndex) -> MigrationMapping: ...

    @overload
    def __getitem__(self, index: slice) -> "MigrationMappingList": ...

    def __getitem__(self, index: SupportsIndex | slice, /) -> "MigrationMapping | MigrationMappingList":
        if isinstance(index, slice):
            return MigrationMappingList(super().__getitem__(index))
        return super().__getitem__(index)

    def get_ids(self) -> list[int]:
        """Return a list of IDs from the migration mappings."""
        return [mapping.id for mapping in self]

    def as_node_ids(self) -> list[NodeId]:
        """Return a list of NodeIds from the migration mappings."""
        return [mapping.instance_id for mapping in self]

    def spaces(self) -> set[str]:
        """Return a set of spaces from the migration mappings."""
        return {mapping.instance_id.space for mapping in self}

    def as_pending_ids(self) -> list[PendingInstanceId]:
        return [PendingInstanceId(pending_instance_id=mapping.instance_id, id=mapping.id) for mapping in self]

    def get_data_set_ids(self) -> set[int]:
        """Return a list of data set IDs from the migration mappings."""
        return {mapping.data_set_id for mapping in self if mapping.data_set_id is not None}

    def as_mapping_by_id(self) -> dict[int, MigrationMapping]:
        """Return a mapping of IDs to MigrationMapping objects."""
        return {mapping.id: mapping for mapping in self}

    @classmethod
    def read_mapping_file(cls, mapping_file: Path, resource_type: str, console: Console | None = None) -> Self:
        # We only validate the schema heading here
        schema = CSVReader.sniff_schema(mapping_file, sniff_rows=1)
        cls._validate_csv_header(schema, console)
        mappings, errors = cls._read_migration_mapping(mapping_file, resource_type)
        return cls(mappings, errors)

    @classmethod
    def _validate_csv_header(cls, schema: list[SchemaColumn], console: Console | None) -> None:
        """Validate the CSV header against the required and optional columns."""
        column_names = {col.name for col in schema}
        missing_required = [col.name for col in cls.REQUIRED_HEADER if col.name not in column_names]
        if missing_required:
            raise ToolkitValueError(
                f"Missing required columns in mapping file: {humanize_collection(missing_required)}."
            )
        expected_names = {col.name for col in cls.REQUIRED_HEADER + cls.OPTIONAL_HEADER}
        ignored = [col.name for col in schema if col.name not in expected_names]
        if ignored:
            LowSeverityWarning(
                f"Ignoring unexpected columns in mapping file: {humanize_collection(ignored)}"
            ).print_warning(console=console)

    @classmethod
    def _read_migration_mapping(
        cls, csv_file: Path, resource_type: str
    ) -> tuple[list[MigrationMapping], dict[int, ValidationError]]:
        """Read a CSV file with ID mappings."""
        mappings: list[MigrationMapping] = []
        errors: dict[int, ValidationError] = {}
        chunk: dict[str, Any]

        def _extract_and_pop(chunk_: dict[str, Any], key_mapping: dict[str, str]) -> dict[str, str]:
            """Pops keys from chunk and returns a new dict with mapped keys."""
            return {dest: chunk_.pop(src) for src, dest in key_mapping.items() if src in chunk_}

        instance_id_mapping = {"space": "space", "externalId": "externalId"}
        consumer_view_mapping = {
            "consumerViewSpace": "space",
            "consumerViewExternalId": "externalId",
            "consumerViewVersion": "version",
        }

        for row_no, chunk in enumerate(CSVReader(csv_file).read_chunks_unprocessed(), 1):
            # Prepare for parsing
            if instance_id := _extract_and_pop(chunk, instance_id_mapping):
                # MyPy does not respect the chunk annotation above, it uses dict[str, str] from the CSVReader.
                chunk["instanceId"] = NodeId.load(instance_id)  # type: ignore[assignment]
            if consumer_view := _extract_and_pop(chunk, consumer_view_mapping):
                chunk["preferredConsumerView"] = ViewId.load(consumer_view)  # type: ignore[assignment]
            chunk["resourceType"] = resource_type
            try:
                mapping = MigrationMapping.model_validate(chunk)
            except ValidationError as e:
                errors[row_no] = e
                continue
            mappings.append(mapping)
        return mappings, errors



from unittest.mock import MagicMock

import pytest
from rich.console import Console

from cognite_toolkit._cdf_tk.utils.text import sanitize_spreadsheet_title, suffix_description


class TestSuffixDescription:
    @pytest.mark.parametrize(
        "description, expected_description, expect_warning",
        [
            (None, "<suffix>", False),
            ("", "<suffix>", False),
            ("Existing description", "Existing description <suffix>", False),
            ("Existing description <suffix>", "Existing description <suffix>", False),
            (
                "A very long description that exceeds the limit with more than one character",
                "A very long descrip...<suffix>",
                True,
            ),
        ],
    )
    def test_suffix_description(self, description: str | None, expected_description: str, expect_warning: True) -> None:
        console = MagicMock(spec=Console)
        limit = 30
        actual = suffix_description("<suffix>", description, limit, "identifier", "my_resource_type", console=console)

        assert actual == expected_description
        assert len(actual) <= limit
        if expect_warning:
            assert console.print.call_count == 1
            message = "".join(console.print.call_args[0])
            assert "Description is too long for my_resource_type 'identifier'." in message


class TestSanitizeSpreadsheetTitle:
    @pytest.mark.parametrize(
        "title, expected",
        [
            ("Valid Title", "Valid Title"),
            ("Invalid/Title", "InvalidTitle"),
            ("Another*Invalid:Title", "AnotherInvalidTitle"),
            ("", "Sheet"),
            (None, "Sheet"),
        ],
    )
    def test_sanitize_spreadsheet_title(self, title: str | None, expected: str) -> None:
        assert sanitize_spreadsheet_title(title) == expected



from abc import ABC, abstractmethod
from collections.abc import Callable, Sequence
from dataclasses import dataclass, field
from typing import Generic, Literal, TypeAlias

import requests

from cognite_toolkit._cdf_tk.utils.http_client._tracker import ItemsRequestTracker
from cognite_toolkit._cdf_tk.utils.useful_types import T_ID, JsonVal

StatusCode: TypeAlias = int


@dataclass
class HTTPMessage(ABC):
    """Base class for HTTP messages (requests and responses)"""

    def dump(self) -> dict[str, JsonVal]:
        """Dumps the message to a JSON serializable dictionary.

        Returns:
            dict[str, JsonVal]: The message as a dictionary.
        """
        # We avoid using the asdict function as we know we have a shallow structure,
        # and this roughly ~10x faster.
        output = self.__dict__.copy()
        output["type"] = type(self).__name__
        return output


@dataclass
class FailedRequestMessage(HTTPMessage):
    error: str


@dataclass
class ResponseMessage(HTTPMessage):
    status_code: StatusCode


@dataclass
class RequestMessage(HTTPMessage):
    """Base class for HTTP request messages"""

    endpoint_url: str
    method: Literal["GET", "POST", "PATCH", "DELETE"]
    connect_attempt: int = 0
    read_attempt: int = 0
    status_attempt: int = 0
    api_version: str | None = None

    @property
    def total_attempts(self) -> int:
        return self.connect_attempt + self.read_attempt + self.status_attempt

    @abstractmethod
    def create_responses(
        self,
        response: requests.Response,
        response_body: dict[str, JsonVal] | None = None,
        error_message: str | None = None,
    ) -> Sequence[HTTPMessage]:
        raise NotImplementedError()

    @abstractmethod
    def create_failed_request(self, error_message: str) -> Sequence[HTTPMessage]:
        raise NotImplementedError()


@dataclass
class SuccessResponse(ResponseMessage):
    body: dict[str, JsonVal] | None = None


@dataclass
class FailedResponse(ResponseMessage):
    error: str
    body: dict[str, JsonVal] | None = None


@dataclass
class SimpleRequest(RequestMessage):
    """Base class for requests with a simple success/fail response structure"""

    @classmethod
    def create_responses(
        cls,
        response: requests.Response,
        response_body: dict[str, JsonVal] | None = None,
        error_message: str | None = None,
    ) -> Sequence[ResponseMessage]:
        if 200 <= response.status_code < 300 and error_message is None:
            return [SuccessResponse(status_code=response.status_code, body=response_body)]
        if error_message is None:
            error_message = f"Request failed with status code {response.status_code}"
        return [FailedResponse(status_code=response.status_code, error=error_message, body=response_body)]

    @classmethod
    def create_failed_request(cls, error_message: str) -> Sequence[HTTPMessage]:
        return [FailedRequestMessage(error=error_message)]


@dataclass
class BodyRequest(RequestMessage, ABC):
    """Base class for HTTP request messages with a body"""

    @abstractmethod
    def body(self) -> dict[str, JsonVal]:
        raise NotImplementedError()


@dataclass
class ParamRequest(SimpleRequest):
    """Base class for HTTP request messages with query parameters"""

    parameters: dict[str, str] | None = None


@dataclass
class SimpleBodyRequest(SimpleRequest, BodyRequest):
    body_content: dict[str, JsonVal] = field(default_factory=dict)

    def body(self) -> dict[str, JsonVal]:
        return self.body_content


@dataclass
class ItemMessage(ABC):
    """Base class for message related to a specific item"""

    ...


@dataclass
class ItemIDMessage(Generic[T_ID], ItemMessage, ABC):
    """Base class for message related to a specific item identified by an ID"""

    id: T_ID


@dataclass
class ItemResponse(ItemIDMessage, ResponseMessage, ABC): ...


@dataclass
class SuccessItem(ItemResponse):
    item: JsonVal | None = None


@dataclass
class FailedItem(ItemResponse):
    error: str


@dataclass
class MissingItem(ItemResponse): ...


@dataclass
class UnexpectedItem(ItemResponse):
    item: JsonVal | None = None


@dataclass
class FailedRequestItem(ItemIDMessage, FailedRequestMessage): ...


@dataclass
class UnknownRequestItem(ItemMessage, FailedRequestMessage):
    item: JsonVal | None = None


@dataclass
class UnknownResponseItem(ItemMessage, ResponseMessage):
    error: str
    item: JsonVal | None = None


@dataclass
class ItemsRequest(Generic[T_ID], BodyRequest):
    """Requests message for endpoints that accept multiple items in a single request.

    This class provides functionality to split large requests into smaller ones, handle responses for each item,
    and manage errors effectively.

    Attributes:
        items (list[JsonVal]): The list of items to be sent in the request body.
        extra_body_fields (dict[str, JsonVal]): Additional fields to include in the request body
        as_id (Callable[[JsonVal], T_ID] | None): A function to extract the ID from each item. If None, IDs are not used.
        max_failures_before_abort (int): The maximum number of failed split requests before aborting further splits.

    """

    items: list[JsonVal] = field(default_factory=list)
    extra_body_fields: dict[str, JsonVal] = field(default_factory=dict)
    as_id: Callable[[JsonVal], T_ID] | None = None
    max_failures_before_abort: int = 50
    tracker: ItemsRequestTracker | None = field(default=None, init=False)

    def dump(self) -> dict[str, JsonVal]:
        """Dumps the message to a JSON serializable dictionary.

        This override removes the 'as_id' attribute as it is not serializable.

        Returns:
            dict[str, JsonVal]: The message as a dictionary.
        """
        output = super().dump()
        if self.as_id is not None:
            # We cannot serialize functions
            del output["as_id"]
        if self.tracker is not None:
            # We cannot serialize the tracker
            del output["tracker"]
        return output

    def body(self) -> dict[str, JsonVal]:
        if self.extra_body_fields:
            return {"items": self.items, **self.extra_body_fields}
        return {"items": self.items}

    def split(self, status_attempts: int) -> "list[ItemsRequest]":
        """Splits the request into two smaller requests.

        This is useful for retrying requests that fail due to size limits or timeouts.

        Args:
            status_attempts: The number of status attempts to set for the new requests. This is used when the
                request failed with a 5xx status code and we want to track the number of attempts. For 4xx errors,
                there is at least one item causing the error, so we do not increment the status attempts, but
                instead essentially do a binary search to find the problematic item(s).

        Returns:
            A list containing two new ItemsRequest instances, each with half of the original items.

        """
        mid = len(self.items) // 2
        if mid == 0:
            return [self]
        tracker = self.tracker or ItemsRequestTracker(self.max_failures_before_abort)
        tracker.register_failure()
        first_half = ItemsRequest[T_ID](
            endpoint_url=self.endpoint_url,
            method=self.method,
            items=self.items[:mid],
            extra_body_fields=self.extra_body_fields,
            as_id=self.as_id,
            connect_attempt=self.connect_attempt,
            read_attempt=self.read_attempt,
            status_attempt=status_attempts,
        )
        first_half.tracker = tracker
        second_half = ItemsRequest[T_ID](
            endpoint_url=self.endpoint_url,
            method=self.method,
            items=self.items[mid:],
            extra_body_fields=self.extra_body_fields,
            as_id=self.as_id,
            connect_attempt=self.connect_attempt,
            read_attempt=self.read_attempt,
            status_attempt=status_attempts,
        )
        second_half.tracker = tracker
        return [first_half, second_half]

    def create_responses(
        self,
        response: requests.Response,
        response_body: dict[str, JsonVal] | None = None,
        error_message: str | None = None,
    ) -> Sequence[HTTPMessage]:
        """Creates response messages based on the HTTP response and the original request.

        Args:
            response: The HTTP response received from the server.
            response_body: The parsed JSON body of the response, if available.
            error_message: An optional error message to use if the response indicates a failure.

        Returns:
            A sequence of HTTPMessage instances representing the outcome for each item in the request.
        """
        if self.as_id is None:
            return SimpleBodyRequest.create_responses(response, response_body, error_message)
        request_items_by_id, errors = self._create_items_by_id()
        responses: list[HTTPMessage] = list(errors)
        error_message = error_message or "Unknown error"

        if not self._is_items_response(response_body):
            return self._handle_non_items_response(responses, response, error_message, request_items_by_id)

        # Process items from response
        if response_body is not None:
            self._process_response_items(responses, response, response_body, error_message, request_items_by_id)

        # Handle missing items
        self._handle_missing_items(responses, response, request_items_by_id)

        return responses

    @staticmethod
    def _handle_non_items_response(
        responses: list[HTTPMessage],
        response: requests.Response,
        error_message: str,
        request_items_by_id: dict[T_ID, JsonVal],
    ) -> list[HTTPMessage]:
        """Handles responses that do not contain an 'items' field in the body."""
        if 200 <= response.status_code < 300:
            responses.extend(
                SuccessItem(status_code=response.status_code, id=id_) for id_ in request_items_by_id.keys()
            )
        else:
            responses.extend(
                FailedItem(status_code=response.status_code, error=error_message, id=id_)
                for id_ in request_items_by_id.keys()
            )
        return responses

    def _process_response_items(
        self,
        responses: list[HTTPMessage],
        response: requests.Response,
        response_body: dict[str, JsonVal],
        error_message: str,
        request_items_by_id: dict[T_ID, JsonVal],
    ) -> None:
        """Processes each item in the response body and categorizes them based on their status."""
        for response_item in response_body["items"]:  # type: ignore[union-attr]
            try:
                item_id = self.as_id(response_item)  # type: ignore[misc]
            except Exception as e:
                responses.append(
                    UnknownResponseItem(
                        status_code=response.status_code, item=response_item, error=f"Error extracting ID: {e!s}"
                    )
                )
                continue
            request_item = request_items_by_id.pop(item_id, None)
            if request_item is None:
                responses.append(UnexpectedItem(status_code=response.status_code, id=item_id, item=response_item))
            elif 200 <= response.status_code < 300:
                responses.append(SuccessItem(status_code=response.status_code, id=item_id, item=response_item))
            else:
                responses.append(FailedItem(status_code=response.status_code, id=item_id, error=error_message))

    @staticmethod
    def _handle_missing_items(
        responses: list[HTTPMessage], response: requests.Response, request_items_by_id: dict[T_ID, JsonVal]
    ) -> None:
        """Handles items that were in the request but not present in the response."""
        for item_id in request_items_by_id.keys():
            responses.append(MissingItem(status_code=response.status_code, id=item_id))

    def create_failed_request(self, error_message: str) -> Sequence[HTTPMessage]:
        if self.as_id is None:
            return SimpleBodyRequest.create_failed_request(error_message)
        items_by_id, errors = self._create_items_by_id()
        results: list[HTTPMessage] = []
        results.extend(errors)
        results.extend(FailedRequestItem(id=item_id, error=error_message) for item_id in items_by_id.keys())
        return results

    def _create_items_by_id(self) -> tuple[dict[T_ID, JsonVal], list[FailedRequestItem | UnknownRequestItem]]:
        if self.as_id is None:
            raise ValueError("as_id function must be provided to create items by ID")
        items_by_id: dict[T_ID, JsonVal] = {}
        errors: list[FailedRequestItem | UnknownRequestItem] = []
        for item in self.items:
            try:
                item_id = self.as_id(item)
            except Exception as e:
                errors.append(UnknownRequestItem(error=f"Error extracting ID: {e!s}", item=item))
                continue
            if item_id in items_by_id:
                errors.append(FailedRequestItem(id=item_id, error=f"Duplicate item ID: {item_id!r}"))
            else:
                items_by_id[item_id] = item
        return items_by_id, errors

    @staticmethod
    def _is_items_response(body: dict[str, JsonVal] | None = None) -> bool:
        if body is None:
            return False
        if "items" not in body:
            return False
        if not isinstance(body["items"], list):
            return False
        return True



import json
from datetime import datetime, timezone
from typing import ClassVar

import pytest
from cognite.client.data_classes import Asset, Event, FileMetadata, Sequence, TimeSeries
from cognite.client.data_classes.data_modeling import NodeId
from cognite.client.data_classes.data_modeling import data_types as dt
from cognite.client.data_classes.data_modeling.data_types import DirectRelationReference, EnumValue
from cognite.client.data_classes.data_modeling.ids import ContainerId, ViewId
from cognite.client.data_classes.data_modeling.views import MappedProperty, MultiEdgeConnection, ViewProperty

from cognite_toolkit._cdf_tk.client.data_classes.migration import AssetCentricId, AssetCentricToViewMapping, ViewSource
from cognite_toolkit._cdf_tk.commands._migrate.conversion import asset_centric_to_dm
from cognite_toolkit._cdf_tk.commands._migrate.issues import (
    ConversionIssue,
    FailedConversion,
    InvalidPropertyDataType,
    ReadAPIIssue,
    ReadFileIssue,
    ReadIssue,
    WriteIssue,
)


class TestMigrationIssues:
    def test_read_file_issue(self) -> None:
        issue = ReadFileIssue(row_no=10, error="Cannot read column 'id' value is not an integer")
        assert issue.dump() == {
            "type": "fileRead",
            "rowNo": 10,
            "error": "Cannot read column 'id' value is not an integer",
        }

    def test_read_api_issue(self) -> None:
        asset_centric_id = AssetCentricId(resource_type="asset", id_=123)
        issue = ReadAPIIssue(asset_centric_id=asset_centric_id, error="API error")
        assert issue.dump() == {
            "type": "apiRead",
            "assetCentricId": {"resourceType": "asset", "id": 123},
            "error": "API error",
        }

    def test_read_issue_subclass(self) -> None:
        class CustomReadIssue(ReadIssue):
            type: ClassVar[str] = "customRead"
            custom_field: str

        issue = CustomReadIssue(custom_field="custom value")
        assert issue.dump() == {
            "type": "customRead",
            "customField": "custom value",
        }

    def test_conversion_issue_minimal(self) -> None:
        asset_centric_id = AssetCentricId(resource_type="asset", id_=456)
        instance_id = NodeId(space="test_space", external_id="test_instance")

        conversion_issue = ConversionIssue(asset_centric_id=asset_centric_id, instance_id=instance_id)

        assert conversion_issue.dump() == {
            "type": "conversion",
            "assetCentricId": {"resourceType": "asset", "id": 456},
            "instanceId": {"space": "test_space", "externalId": "test_instance", "type": "node"},
            "failedConversions": [],
            "invalidInstancePropertyTypes": [],
            "missingAssetCentricProperties": [],
            "missingInstanceProperties": [],
            "ignoredAssetCentricProperties": [],
        }

    def test_conversion_issue_with_all_fields(self) -> None:
        asset_centric_id = AssetCentricId(resource_type="timeseries", id_=789)
        instance_id = NodeId(space="demo_space", external_id="demo_instance")

        failed_conversion = FailedConversion(property_id="value", value="not_a_number", error="Cannot convert to float")
        invalid_property = InvalidPropertyDataType(property_id="status", expected_type="boolean")

        conversion_issue = ConversionIssue(
            asset_centric_id=asset_centric_id,
            instance_id=instance_id,
            missing_asset_centric_properties=["missing_source_prop"],
            missing_instance_properties=["missing_target_prop"],
            invalid_instance_property_types=[invalid_property],
            failed_conversions=[failed_conversion],
        )

        assert conversion_issue.dump() == {
            "type": "conversion",
            "assetCentricId": {"resourceType": "timeseries", "id": 789},
            "instanceId": {"space": "demo_space", "externalId": "demo_instance", "type": "node"},
            "missingAssetCentricProperties": ["missing_source_prop"],
            "missingInstanceProperties": ["missing_target_prop"],
            "invalidInstancePropertyTypes": [{"propertyId": "status", "expectedType": "boolean"}],
            "failedConversions": [{"propertyId": "value", "value": "not_a_number", "error": "Cannot convert to float"}],
            "ignoredAssetCentricProperties": [],
        }

    def test_write_issue_minimal(self) -> None:
        instance_id = NodeId(space="write_space", external_id="write_instance")

        write_issue = WriteIssue(instance_id=instance_id, status_code=400)

        assert write_issue.dump() == {
            "type": "write",
            "instanceId": {"space": "write_space", "externalId": "write_instance", "type": "node"},
            "statusCode": 400,
            "message": None,
        }

    def test_write_issue_with_message(self) -> None:
        instance_id = NodeId(space="error_space", external_id="error_instance")

        write_issue = WriteIssue(instance_id=instance_id, status_code=500, message="Internal server error occurred")

        assert write_issue.dump() == {
            "type": "write",
            "instanceId": {"space": "error_space", "externalId": "error_instance", "type": "node"},
            "statusCode": 500,
            "message": "Internal server error occurred",
        }


class TestAssetCentricConversion:
    INSTANCE_ID = NodeId(space="test_space", external_id="test_instance")
    CONTAINER_ID = ContainerId("test_space", "test_container")
    VIEW_ID = ViewId("test_space", "test_view", "v1")
    INSTANCE_SOURCE_VIEW_ID = ViewId("cognite_migration", "InstanceSource", "v1")

    @pytest.mark.parametrize(
        "resource,view_source,view_properties,expected_properties,expected_issue",
        [
            pytest.param(
                # Simple Asset with basic mapping
                Asset(id=123, external_id="asset_123", name="Test Asset", description="A test asset"),
                ViewSource(
                    external_id="asset_mapping",
                    version=1,
                    last_updated_time=1000000,
                    created_time=1000000,
                    resource_type="asset",
                    view_id=ViewId("test_space", "test_view", "v1"),
                    mapping=AssetCentricToViewMapping(
                        to_property_id={"name": "assetName", "description": "assetDescription"}
                    ),
                ),
                {
                    "assetName": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "assetName",
                        dt.Text(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                    "assetDescription": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "assetDescription",
                        dt.Text(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                },
                {"assetName": "Test Asset", "assetDescription": "A test asset"},
                ConversionIssue(
                    asset_centric_id=AssetCentricId("asset", 123),
                    instance_id=INSTANCE_ID,
                    ignored_asset_centric_properties=[],
                ),
                id="simple_asset_mapping",
            ),
            pytest.param(
                TimeSeries(
                    id=456,
                    external_id="ts_456",
                    name="Test TimeSeries",
                    description="A test timeseries",
                    unit="celsius",
                    metadata={"sensor_type": "temperature", "location": "room_1"},
                ),
                ViewSource(
                    external_id="timeseries_mapping",
                    version=1,
                    last_updated_time=1000000,
                    created_time=1000000,
                    resource_type="timeseries",
                    view_id=ViewId("test_space", "test_view", "v1"),
                    mapping=AssetCentricToViewMapping(
                        to_property_id={"name": "timeseriesName", "unit": "measurementUnit"},
                        metadata_to_property_id={"sensor_type": "sensorType", "location": "deviceLocation"},
                    ),
                ),
                {
                    "timeseriesName": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "timeseriesName",
                        dt.Text(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                    "measurementUnit": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "measurementUnit",
                        dt.Text(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                    "sensorType": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "sensorType",
                        dt.Text(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                    "deviceLocation": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "deviceLocation",
                        dt.Text(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                },
                {
                    "timeseriesName": "Test TimeSeries",
                    "measurementUnit": "celsius",
                    "sensorType": "temperature",
                    "deviceLocation": "room_1",
                },
                ConversionIssue(
                    asset_centric_id=AssetCentricId("timeseries", id_=456),
                    instance_id=INSTANCE_ID,
                    ignored_asset_centric_properties=["description"],
                ),
                id="timeseries_with_metadata",
            ),
            pytest.param(
                Event(
                    id=789,
                    external_id="event_789",
                    start_time=1756359489386,
                    end_time=1756359499880,
                    description="Not mapped",
                    metadata={
                        "operator": "John Doe",
                        "severity": "HIGH",
                        "value": "invalid_int",  # This will cause a conversion issue
                        "aConnectionProp": json.dumps(
                            {"externalId": "op_123", "space": "schema_space", "type": "Operation"}
                        ),
                    },
                ),
                ViewSource(
                    external_id="incomplete_mapping",
                    version=1,
                    last_updated_time=1000000,
                    created_time=1000000,
                    resource_type="event",
                    view_id=ViewId("test_space", "test_view", "v1"),
                    mapping=AssetCentricToViewMapping(
                        to_property_id={
                            "missing_prop": "targetProp",
                            "startTime": "eventStart",
                            "endTime": "eventEnd",
                        },
                        metadata_to_property_id={
                            "operator": "missingDMProp",
                            "severity": "eventSeverity",
                            "value": "eventValue",
                            "missingMetaProp": "anotherMissingDMProp",
                            "aConnectionProp": "some_other_event",
                        },
                    ),
                ),
                {
                    "eventStart": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "eventStart",
                        dt.Timestamp(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                    "eventEnd": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "eventEnd",
                        dt.Timestamp(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                    "eventSeverity": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "eventSeverity",
                        dt.Enum({"high": EnumValue(), "low": EnumValue(), "medium": EnumValue()}),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                    "eventValue": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "eventValue",
                        dt.Int64(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                    "some_other_event": MultiEdgeConnection(
                        type=DirectRelationReference("schema_space", "Operation"),
                        source=ViewId("test_space", "test_view", "v1"),
                        name=None,
                        description=None,
                        edge_source=None,
                        direction="outwards",
                    ),
                },
                {
                    "eventStart": datetime(2025, 8, 28, 5, 38, 9, 386000, tzinfo=timezone.utc),
                    "eventEnd": datetime(2025, 8, 28, 5, 38, 19, 880000, tzinfo=timezone.utc),
                    "eventSeverity": "high",
                },
                ConversionIssue(
                    asset_centric_id=AssetCentricId("event", id_=789),
                    instance_id=INSTANCE_ID,
                    ignored_asset_centric_properties=["description"],
                    missing_asset_centric_properties=["missing_prop", "metadata.missingMetaProp"],
                    missing_instance_properties=["missingDMProp"],
                    invalid_instance_property_types=[
                        InvalidPropertyDataType(property_id="some_other_event", expected_type="MappedProperty")
                    ],
                    failed_conversions=[
                        FailedConversion(
                            property_id="value",
                            value="invalid_int",
                            error="Cannot convert invalid_int to int64.",
                        )
                    ],
                ),
                id="Event with conversion issues (missing properties)",
            ),
            pytest.param(
                FileMetadata(
                    id=321,
                    external_id="file_321",
                    name="Test File",
                    mime_type="application/octet-stream",
                    metadata={"file_type": "pdf", "confidential": "true"},
                ),
                ViewSource(
                    external_id="file_mapping",
                    version=1,
                    last_updated_time=1000000,
                    created_time=1000000,
                    resource_type="file",
                    view_id=ViewId("test_space", "test_view", "v1"),
                    mapping=AssetCentricToViewMapping(
                        to_property_id={},
                        metadata_to_property_id={},
                    ),
                ),
                {
                    "fileName": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "fileName",
                        dt.Text(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                    "fileDescription": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "fileDescription",
                        dt.Text(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                    "fileType": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "fileType",
                        dt.Text(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                    "isConfidential": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "isConfidential",
                        dt.Boolean(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                },
                {},
                ConversionIssue(
                    asset_centric_id=AssetCentricId("file", id_=321),
                    instance_id=INSTANCE_ID,
                    ignored_asset_centric_properties=[
                        "metadata.confidential",
                        "metadata.file_type",
                        "mimeType",
                        "name",
                    ],
                    missing_asset_centric_properties=[],
                    missing_instance_properties=[],
                ),
                id="FileMetadata with no mappings (all ignored)",
            ),
            pytest.param(
                Sequence(
                    id=654,
                    name="Test Sequence",
                    description="A test sequence",
                    metadata=None,
                ),
                ViewSource(
                    external_id="sequence_mapping",
                    version=1,
                    last_updated_time=1000000,
                    created_time=1000000,
                    resource_type="sequence",
                    view_id=ViewId("test_space", "test_view", "v1"),
                    mapping=AssetCentricToViewMapping(
                        to_property_id={"name": "sequenceName"},
                        metadata_to_property_id={"category": "sequenceCategory"},
                    ),
                ),
                {
                    "sequenceName": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "sequenceName",
                        dt.Text(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                    "sequenceCategory": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "sequenceCategory",
                        dt.Text(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                },
                {"sequenceName": "Test Sequence"},
                ConversionIssue(
                    asset_centric_id=AssetCentricId("sequence", id_=654),
                    instance_id=INSTANCE_ID,
                    ignored_asset_centric_properties=["description"],
                    missing_asset_centric_properties=["metadata.category"],
                    missing_instance_properties=[],
                ),
                id="Sequence with partial mapping",
            ),
            pytest.param(
                Asset(id=999, external_id="asset_999", name=None, description=None),
                ViewSource(
                    external_id="empty_mapping",
                    version=1,
                    last_updated_time=1000000,
                    created_time=1000000,
                    resource_type="asset",
                    view_id=ViewId("test_space", "test_view", "v1"),
                    mapping=AssetCentricToViewMapping(
                        to_property_id={"name": "assetName", "description": "assetDescription"}
                    ),
                ),
                {
                    "assetName": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "assetName",
                        dt.Text(),
                        nullable=False,
                        immutable=False,
                        auto_increment=False,
                    ),
                    "assetDescription": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "assetDescription",
                        dt.Text(),
                        nullable=False,
                        immutable=False,
                        auto_increment=False,
                    ),
                },
                {},
                ConversionIssue(
                    asset_centric_id=AssetCentricId("asset", id_=999),
                    instance_id=INSTANCE_ID,
                    ignored_asset_centric_properties=[],
                    # Name and description set to None is the same as missing as we have now way of knowing
                    # whether they were explicitly set to None or just not set at all.
                    missing_asset_centric_properties=["name", "description"],
                    missing_instance_properties=[],
                ),
                id="Asset with non-nullable properties all None",
            ),
            pytest.param(
                Event(id=999, external_id="event_999", type="MyType", metadata={"category": "MyCategory"}),
                ViewSource(
                    external_id="event_mapping",
                    version=1,
                    last_updated_time=1000000,
                    created_time=1000000,
                    resource_type="event",
                    view_id=ViewId("test_space", "test_view", "v1"),
                    mapping=AssetCentricToViewMapping(
                        to_property_id={"type": "category"},
                        metadata_to_property_id={"category": "category"},
                    ),
                ),
                {
                    "category": MappedProperty(
                        ContainerId("test_space", "test_container"),
                        "category",
                        dt.Text(),
                        nullable=True,
                        immutable=False,
                        auto_increment=False,
                    ),
                },
                {
                    "category": "MyType",
                },
                ConversionIssue(
                    asset_centric_id=AssetCentricId("event", id_=999),
                    instance_id=INSTANCE_ID,
                    ignored_asset_centric_properties=["metadata.category"],
                    missing_asset_centric_properties=[],
                    missing_instance_properties=[],
                ),
                id="Event with overlapping property and metadata mapping (property takes precedence)",
            ),
        ],
    )
    def test_asset_centric_to_dm(
        self,
        resource: Asset | FileMetadata | Event | TimeSeries | Sequence,
        view_source: ViewSource,
        view_properties: dict[str, ViewProperty],
        expected_properties: dict[str, str],
        expected_issue: ConversionIssue,
    ) -> None:
        actual, issue = asset_centric_to_dm(resource, self.INSTANCE_ID, view_source, view_properties)

        # Check the structure of the returned NodeApply
        assert actual.space == self.INSTANCE_ID.space
        assert actual.external_id == self.INSTANCE_ID.external_id
        assert 1 <= len(actual.sources) <= 2

        # Check the main view source
        if len(actual.sources) == 2:
            main_source = actual.sources[0]
            assert main_source.source == view_source.view_id
            assert main_source.properties == expected_properties
        else:
            assert expected_properties == {}

        # Check the instance source view
        instance_source = actual.sources[-1]
        assert instance_source.source == self.INSTANCE_SOURCE_VIEW_ID
        assert instance_source.properties["resourceType"] == view_source.resource_type
        assert instance_source.properties["id"] == resource.id
        assert instance_source.properties["dataSetId"] == resource.data_set_id
        assert instance_source.properties["classicExternalId"] == resource.external_id

        assert expected_issue.dump() == issue.dump()



from collections.abc import Iterable
from datetime import datetime
from pathlib import Path

import pytest
from cognite.client.data_classes import FileMetadataWrite, TimeSeriesWrite
from cognite.client.data_classes.data_modeling import NodeId, Space
from cognite.client.data_classes.data_modeling.cdm.v1 import CogniteFileApply, CogniteTimeSeriesApply
from cognite.client.utils import datetime_to_ms

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.commands import PurgeCommand
from cognite_toolkit._cdf_tk.storageio import InstanceFileSelector
from tests.test_integration.constants import RUN_UNIQUE_ID


@pytest.fixture()
def file_ts_nodes(
    toolkit_client_with_pending_ids: ToolkitClient, toolkit_space: Space
) -> Iterable[tuple[tuple[NodeId, int], tuple[NodeId, int]]]:
    client = toolkit_client_with_pending_ids
    file = CogniteFileApply(
        space=toolkit_space.space,
        external_id=f"test_file_purge_with_unlink_{RUN_UNIQUE_ID}",
        name="Test File for Purge with Unlink",
        mime_type="text/plain",
    )
    ts = CogniteTimeSeriesApply(
        space=toolkit_space.space,
        external_id=f"test_ts_purge_with_unlink_{RUN_UNIQUE_ID}",
        name="Test TS for Purge with Unlink",
        is_step=False,
        time_series_type="numeric",
    )
    classic_file = FileMetadataWrite(
        name=file.name,
        external_id=file.external_id,
        mime_type=file.mime_type,
    )
    classic_ts = TimeSeriesWrite(
        external_id=ts.external_id,
        name=ts.name,
        is_step=ts.is_step,
        is_string=ts.time_series_type == "string",
    )
    file_id: int | None = None
    ts_id: int | None = None
    try:
        # Ensure clean state
        client.data_modeling.instances.delete([file.as_id(), ts.as_id()])
        client.files.delete(external_id=classic_file.external_id, ignore_unknown_ids=True)
        client.time_series.delete(external_id=classic_ts.external_id, ignore_unknown_ids=True)

        # Create timeseries and file
        created_file = client.files.upload_bytes(b"Sample file content", **classic_file.dump(camel_case=False))
        file_id = created_file.id
        created_ts = client.time_series.create(classic_ts)
        ts_id = created_ts.id
        client.time_series.data.insert(
            datapoints=[{"timestamp": datetime_to_ms(datetime(2020, 1, 1, 0, 0, 0)), "value": 1.0}],
            id=ts_id,
        )

        # Link them.
        client.files.set_pending_ids(file.as_id(), id=file_id)
        client.time_series.set_pending_ids(ts.as_id(), id=ts_id)

        # Create Nodes in CDM
        created = client.data_modeling.instances.apply([file, ts])
        assert len(created.nodes) == 2

        yield (file.as_id(), file_id), (ts.as_id(), ts_id)
    finally:
        client.data_modeling.instances.delete([file.as_id(), ts.as_id()])
        if file_id is not None:
            client.files.unlink_instance_ids(id=file_id)
            client.files.delete(id=file_id, ignore_unknown_ids=True)
        if ts_id is not None:
            client.time_series.unlink_instance_ids(id=ts_id)
            client.time_series.delete(id=ts_id, ignore_unknown_ids=True)


class TestPurge:
    def test_purge_instances_with_unlink(
        self,
        file_ts_nodes: tuple[tuple[NodeId, int], tuple[NodeId, int]],
        toolkit_client_with_pending_ids: ToolkitClient,
        tmp_path: Path,
    ) -> None:
        client = toolkit_client_with_pending_ids
        (file_node, file_id), (ts_node, ts_id) = file_ts_nodes

        csv_path = tmp_path / "test.csv"
        csv_path.write_text(
            f"""space,externalId,instanceType
{file_node.space},{file_node.external_id},node
{ts_node.space},{ts_node.external_id},node
""",
            encoding="utf-8",
        )

        purge = PurgeCommand(silent=True)

        results = purge.instances(
            client,
            InstanceFileSelector(datafile=csv_path, validate=True),
            dry_run=False,
            unlink=True,
            verbose=False,
            auto_yes=True,
        )
        assert results.deleted == 2

        results = client.data_modeling.instances.retrieve([file_node, ts_node])
        assert len(results.nodes) == 0, "Instances were not purged"

        classic_file = client.files.retrieve(id=file_id)
        assert classic_file is not None, "File was not unlinked"

        classic_ts = client.time_series.retrieve(id=ts_id)
        assert classic_ts is not None, "Time series was not unlinked"



from abc import ABC, abstractmethod
from typing import Generic

from cognite.client.data_classes._base import (
    T_CogniteResourceList,
)
from cognite.client.data_classes.data_modeling import View, ViewId

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.instances import InstanceApplyList
from cognite_toolkit._cdf_tk.client.data_classes.migration import ViewSource
from cognite_toolkit._cdf_tk.commands._migrate.adapter import (
    AssetCentricMappingList,
    MigrationSelector,
)
from cognite_toolkit._cdf_tk.commands._migrate.conversion import asset_centric_to_dm
from cognite_toolkit._cdf_tk.commands._migrate.issues import MigrationIssue
from cognite_toolkit._cdf_tk.exceptions import ToolkitValueError
from cognite_toolkit._cdf_tk.storageio._base import T_Selector, T_WritableCogniteResourceList
from cognite_toolkit._cdf_tk.utils import humanize_collection


class DataMapper(Generic[T_Selector, T_WritableCogniteResourceList, T_CogniteResourceList], ABC):
    def prepare(self, source_selector: T_Selector) -> None:
        """Prepare the data mapper with the given source selector.

        Args:
            source_selector: The selector for the source data.

        """
        # Override in subclass if needed.
        pass

    @abstractmethod
    def map_chunk(self, source: T_WritableCogniteResourceList) -> tuple[T_CogniteResourceList, list[MigrationIssue]]:
        """Map a chunk of source data to the target format.

        Args:
            source: The source data chunk to be mapped.

        Returns:
            A tuple containing the mapped data and a list of any issues encountered during mapping.

        """
        raise NotImplementedError("Subclasses must implement this method.")


class AssetCentricMapper(DataMapper[MigrationSelector, AssetCentricMappingList, InstanceApplyList]):
    def __init__(self, client: ToolkitClient) -> None:
        self.client = client
        self._ingestion_view_by_id: dict[ViewId, View] = {}
        self._view_mapping_by_id: dict[str, ViewSource] = {}

    def prepare(self, source_selector: MigrationSelector) -> None:
        ingestion_view_ids = source_selector.get_ingestion_views()
        ingestion_views = self.client.migration.view_source.retrieve(ingestion_view_ids)
        self._view_mapping_by_id = {view.external_id: view for view in ingestion_views}
        missing_mappings = set(ingestion_view_ids) - set(self._view_mapping_by_id.keys())
        if missing_mappings:
            raise ToolkitValueError(
                f"The following ingestion views were not found: {humanize_collection(missing_mappings)}"
            )

        view_ids = list({view.view_id for view in ingestion_views})
        views = self.client.data_modeling.views.retrieve(view_ids)
        self._ingestion_view_by_id = {view.as_id(): view for view in views}
        missing_views = set(view_ids) - set(self._ingestion_view_by_id.keys())
        if missing_views:
            raise ToolkitValueError(
                f"The following ingestion views were not found in Data Modeling: {humanize_collection(missing_views)}"
            )

    def map_chunk(self, source: AssetCentricMappingList) -> tuple[InstanceApplyList, list[MigrationIssue]]:
        """Map a chunk of asset-centric data to InstanceApplyList format."""
        instances = InstanceApplyList([])
        issues: list[MigrationIssue] = []
        for item in source:
            mapping = item.mapping
            ingestion_view = mapping.get_ingestion_view()
            try:
                view_source = self._view_mapping_by_id[ingestion_view]
                view_properties = self._ingestion_view_by_id[view_source.view_id].properties
            except KeyError as e:
                raise RuntimeError(
                    f"Failed to lookup mapping or view for ingestion view '{ingestion_view}'. Did you forget to call .prepare()?"
                ) from e
            instance, conversion_issue = asset_centric_to_dm(
                item.resource,
                instance_id=mapping.instance_id,
                view_source=view_source,
                view_properties=view_properties,
            )

            instances.append(instance)
            if conversion_issue.has_issues:
                issues.append(conversion_issue)
        return instances, issues



from abc import ABC, abstractmethod
from collections.abc import Iterable, Mapping, Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import ClassVar, Generic

from cognite.client.data_classes._base import (
    T_CogniteResourceList,
)
from rich.console import Console

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.exceptions import ToolkitNotImplementedError
from cognite_toolkit._cdf_tk.utils.collection import chunker_sequence
from cognite_toolkit._cdf_tk.utils.fileio import SchemaColumn
from cognite_toolkit._cdf_tk.utils.http_client import HTTPClient, HTTPMessage, ItemsRequest
from cognite_toolkit._cdf_tk.utils.useful_types import T_ID, JsonVal, T_Selector, T_WritableCogniteResourceList


@dataclass
class StorageIOConfig:
    kind: str
    folder_name: str
    value: JsonVal


class StorageIO(ABC, Generic[T_ID, T_Selector, T_CogniteResourceList, T_WritableCogniteResourceList]):
    """This is a base class for all storage classes in Cognite Toolkit

    It defines the interface for interacting with storage items in CDF, such as downloading,
    uploading, and managing configurations. Each storage type (e.g., RAW, asset-centric)
    should implement this interface to provide specific functionality.

    Attributes:
        folder_name: The name of the folder in which the storage item is located.
        kind: The type of storage (e.g., 'raw', 'assets').
        display_name: A human-readable name for the storage item.
        supported_download_formats: A set of formats that the storage item supports for downloading.
        supported_compressions: A set of compression formats that the storage item supports.
        supported_read_formats: A set of formats that the storage item supports for reading.
        chunk_size: The size of the data chunks to be processed during download and upload operations.
        client: An instance of ToolkitClient to interact with the CDF API.
    """

    folder_name: str
    kind: str
    display_name: str
    supported_download_formats: frozenset[str]
    supported_compressions: frozenset[str]
    supported_read_formats: frozenset[str]
    chunk_size: int
    UPLOAD_ENDPOINT: ClassVar[str]
    UPLOAD_EXTRA_ARGS: ClassVar[Mapping[str, JsonVal] | None] = None

    def __init__(self, client: ToolkitClient) -> None:
        self.client = client

    @abstractmethod
    def as_id(self, item: dict[str, JsonVal] | object) -> T_ID:
        """Convert an item to its corresponding ID.
        Args:
            item: The item to convert.
        Returns:
            The ID corresponding to the item.
        """
        raise NotImplementedError()

    @abstractmethod
    def download_iterable(
        self, selector: T_Selector, limit: int | None = None
    ) -> Iterable[T_WritableCogniteResourceList]:
        """Download items from the storage given the selection criteria.

        Args:
            selector: The selection criteria to filter the items to download.
            limit: Optional limit on the number of items to download.

        Returns:
            An iterable of writable Cognite resource lists.
        """
        raise NotImplementedError()

    @abstractmethod
    def count(self, selector: T_Selector) -> int | None:
        """Count the number of items in the storage that match the given selector.

        Args:
            selector: The selection criteria to filter the items to count.

        Returns:
            The number of items that match the selection criteria, or None if counting is not supported.
        """
        raise NotImplementedError()

    @abstractmethod
    def upload_items(self, data_chunk: T_CogniteResourceList, selector: T_Selector) -> None:
        """Upload a chunk of data to the storage.

        Args:
            data_chunk: The chunk of data to upload, which should be a list of writable Cognite resources.
            selector: The selection criteria to identify where to upload the data.
        """
        raise NotImplementedError()

    def upload_items_force(
        self, data_chunk: T_CogniteResourceList, http_client: HTTPClient, selector: T_Selector | None = None
    ) -> Sequence[HTTPMessage]:
        """Upload a chunk of data to the storage using a custom HTTP client.
        This ensures that even if one item in the chunk fails, the rest will still be uploaded.

        Args:
            data_chunk: The chunk of data to upload, which should be a list of writable Cognite resources.
            http_client: The custom HTTP client to use for the upload.
            selector: Optional selection criteria to identify where to upload the data.
        """
        if not hasattr(self, "UPLOAD_ENDPOINT"):
            raise ToolkitNotImplementedError(f"Upload not implemented for {self.kind} storage.")

        config = http_client.config
        results: list[HTTPMessage] = []
        for batch in chunker_sequence(data_chunk, self.chunk_size):
            batch_results = http_client.request_with_retries(
                message=ItemsRequest(
                    endpoint_url=config.create_api_url(self.UPLOAD_ENDPOINT),
                    method="POST",
                    # The dump method from the PySDK always returns JsonVal, but mypy cannot infer that
                    items=batch.dump(camel_case=True),  # type: ignore[arg-type]
                    extra_body_fields=dict(self.UPLOAD_EXTRA_ARGS or {}),
                    as_id=self.as_id,
                )
            )
            results.extend(batch_results)
        return results

    @abstractmethod
    def data_to_json_chunk(self, data_chunk: T_WritableCogniteResourceList) -> list[dict[str, JsonVal]]:
        """Convert a chunk of data to a JSON-compatible format.

        Args:
            data_chunk: The chunk of data to convert, which should be a writable Cognite resource list.

        Returns:
            A list of dictionaries representing the data in a JSON-compatible format.

        """
        raise NotImplementedError()

    @abstractmethod
    def json_chunk_to_data(self, data_chunk: list[dict[str, JsonVal]]) -> T_CogniteResourceList:
        """Convert a JSON-compatible chunk of data back to a writable Cognite resource list.

        Args:
            data_chunk: A list of dictionaries representing the data in a JSON-compatible format.
        Returns:
            A writable Cognite resource list representing the data.
        """
        raise NotImplementedError()

    @abstractmethod
    def configurations(self, selector: T_Selector) -> Iterable[StorageIOConfig]:
        """Return configurations for the storage item."""
        raise NotImplementedError()

    @abstractmethod
    def load_selector(self, datafile: Path) -> T_Selector:
        """Load the selector from adjacent filepath."""
        raise NotImplementedError()

    @abstractmethod
    def ensure_configurations(self, selector: T_Selector, console: Console | None = None) -> None:
        """Ensure that the necessary configurations for the storage item are in place.

        This method should create the necessary configurations in CDF if they do not exist.
        For example, for RAW tables, this will create the RAW database and table.

        For asset-centric storage, this will create labels and data sets.

        Args:
            selector: The selection criteria to find the data.
            console: An optional console for outputting messages during the configuration process.

        """
        raise NotImplementedError()


class TableStorageIO(StorageIO[T_ID, T_Selector, T_CogniteResourceList, T_WritableCogniteResourceList], ABC):
    @abstractmethod
    def get_schema(self, selector: T_Selector) -> list[SchemaColumn]:
        """Get the schema of the table associated with the given selector.

        Args:
            selector: The selection criteria to identify the data.

        Returns:
            A list of SchemaColumn objects representing the schema of the table.

        """



from collections.abc import Sequence
from typing import Literal

from cognite.client.data_classes.capabilities import (
    AssetsAcl,
    Capability,
    DataModelInstancesAcl,
    DataModelsAcl,
    FilesAcl,
    TimeSeriesAcl,
)

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.exceptions import AuthorizationError
from cognite_toolkit._cdf_tk.utils import humanize_collection


class ValidateAccess:
    def __init__(self, client: ToolkitClient, default_operation: str) -> None:
        self.client = client
        self.default_operation = default_operation

    def data_model(
        self, action: Sequence[Literal["read", "write"]], spaces: set[str] | None = None, operation: str | None = None
    ) -> list[str] | None:
        """Validate access to data models.

        Args:
            action (Sequence[Literal["read", "write"]]): The actions to validate access for.
            spaces (Set[str] | None): The space IDs to check access for. If None, checks access for all spaces.
            operation (str | None): The operation being performed, used for error messages.

        Returns:
            list[str] | None: Returns a list of space IDs if access is limited to these spaces, or None if access is granted to all spaces.

        Raises:
            ValueError: If the client.token.get_scope() returns an unexpected number of data model scopes.
            AuthorizationError: If the user does not have permission to perform the specified action on the given space.
        """
        operation = operation or self.default_operation
        model_scopes, actions_str = self._set_up_read_write(
            action, DataModelsAcl.Action.Read, DataModelsAcl.Action.Write, operation, "data models"
        )
        if len(model_scopes) != 1:
            raise ValueError(f"Unexpected number of data model scopes: {len(model_scopes)}. Expected 1 scope.")
        model_scope = model_scopes[0]
        if isinstance(model_scope, DataModelsAcl.Scope.All):
            return None
        elif isinstance(model_scope, DataModelsAcl.Scope.SpaceID):
            if spaces is None:
                return model_scope.space_ids
            if missing := spaces - set(model_scope.space_ids):
                raise AuthorizationError(
                    f"You have no permission to {actions_str} the {humanize_collection(missing)!r} space(s). This is required to {operation}."
                )
            return None
        else:
            raise ValueError(f"Unexpected data model scope type: {type(model_scope)}. Expected SpaceID or All.")

    def instances(
        self, action: Sequence[Literal["read", "write"]], spaces: set[str] | None = None, operation: str | None = None
    ) -> list[str] | None:
        """Validate access to data model instances.

        Args:
            action (Sequence[Literal["read", "write"]]): The actions to validate access for.
            spaces (Set[str] | None): The space IDs to check access for. If None, checks access for all spaces.
            operation (str | None): The operation being performed, used for error messages.

        Returns:
            list[str] | None: Returns a list of space IDs if access is limited to these spaces, or None if access is granted to all spaces.

        Raises:
            ValueError: If the client.token.get_scope() returns an unexpected number of data model instance scopes.
            AuthorizationError: If the user does not have permission to perform the specified action on the given space.
        """
        operation = operation or self.default_operation
        instance_scopes, action_str = self._set_up_read_write(
            action, DataModelInstancesAcl.Action.Read, DataModelInstancesAcl.Action.Write, operation, "instances"
        )
        if len(instance_scopes) != 1:
            raise ValueError(
                f"Unexpected number of data model instance scopes: {len(instance_scopes)}. Expected 1 scope."
            )
        instance_scope = instance_scopes[0]
        if isinstance(instance_scope, DataModelInstancesAcl.Scope.SpaceID):
            if spaces is None:
                return instance_scope.space_ids
            if missing := spaces - set(instance_scope.space_ids):
                raise AuthorizationError(
                    f"You have no permission to {action_str} instances in the {humanize_collection(missing)!r} space(s). This is required to {operation} instances."
                )
            return None
        elif isinstance(instance_scope, DataModelInstancesAcl.Scope.All):
            return None
        else:
            raise ValueError(
                f"Unexpected data model instance scope type: {type(instance_scope)}. Expected SpaceID or All."
            )

    def timeseries(
        self,
        action: Sequence[Literal["read", "write"]],
        dataset_ids: set[int] | None = None,
        operation: str | None = None,
    ) -> dict[str, list[str]] | None:
        """Validate access to time series.
        Args:
            action (Sequence[Literal["read", "write"]]): The actions to validate access for.
            dataset_ids (Set[int] | None): The dataset IDs to check access for. If None, checks access for all datasets.
            operation (str | None): The operation being performed, used for error messages.
        Returns:
            dict[str, list[str]] | None: Returns a dictionary with keys 'dataset', 'asset root', and 'time series' if access is limited to these scopes, or None if access is granted to all time series.
        Raises:
            ValueError: If the client.token.get_scope() returns an unexpected timeseries scope type.
            AuthorizationError: If the user does not have permission to perform the specified action on the given dataset or time series.
        """
        operation = operation or self.default_operation
        timeseries_scopes, actions_str = self._set_up_read_write(
            action, TimeSeriesAcl.Action.Read, TimeSeriesAcl.Action.Write, operation, "time series"
        )

        if isinstance(timeseries_scopes[0], TimeSeriesAcl.Scope.All):
            return None
        if dataset_ids is not None:
            missing = set(dataset_ids)
            for scope in timeseries_scopes:
                if isinstance(scope, TimeSeriesAcl.Scope.DataSet):
                    missing = missing - set(scope.ids)
                    if not missing:
                        return None
            raise AuthorizationError(
                f"You have no permission to {actions_str} time series in dataset {humanize_collection(missing)}. This is required to {operation}."
            )
        output: dict[str, list[str]] = {}
        for scope in timeseries_scopes:
            if isinstance(scope, TimeSeriesAcl.Scope.DataSet):
                output["dataset"] = self.client.lookup.data_sets.external_id(scope.ids)
            elif isinstance(scope, TimeSeriesAcl.Scope.AssetRootID):
                output["asset root"] = self.client.lookup.assets.external_id(scope.root_ids)
            elif isinstance(scope, TimeSeriesAcl.Scope.ID):
                output["time series"] = self.client.lookup.time_series.external_id(scope.ids)
            else:
                raise ValueError(
                    f"Unexpected timeseries scope type: {type(scope)}. Expected DataSet, AssetRootID or ID."
                )

        return output

    def files(
        self,
        action: Sequence[Literal["read", "write"]],
        dataset_ids: set[int] | None = None,
        operation: str | None = None,
    ) -> dict[str, list[str]] | None:
        """Validate access to files.

        Args:
            action (Sequence[Literal["read", "write"]]): The actions to validate access for
            dataset_ids (Set[int] | None): The dataset IDs to check access for. If None, checks access for all datasets.
            operation (str | None): The operation being performed, used for error messages.
        Returns:
            dict[str, list[str]] | None: Returns a dictionary with the key 'dataset' if access is limited to a dataset scope, or None if access is granted to all files.
        Raises:
            ValueError: If the client.token.get_scope() returns an unexpected file scope type.
            AuthorizationError: If the user does not have permission to perform the specified action on the given
                dataset.
        """
        operation = operation or self.default_operation
        file_scopes, actions_str = self._set_up_read_write(
            action, FilesAcl.Action.Read, FilesAcl.Action.Write, operation, "files"
        )
        if isinstance(file_scopes[0], FilesAcl.Scope.All):
            return None
        if dataset_ids is not None:
            missing = set(dataset_ids)
            for scope in file_scopes:
                if isinstance(scope, FilesAcl.Scope.DataSet):
                    missing = missing - set(scope.ids)
                    if not missing:
                        return None
            raise AuthorizationError(
                f"You have no permission to {actions_str} files in dataset {humanize_collection(missing)}. This is required to {operation}."
            )
        output: dict[str, list[str]] = {}
        for scope in file_scopes:
            if isinstance(scope, FilesAcl.Scope.DataSet):
                output["dataset"] = self.client.lookup.data_sets.external_id(scope.ids)
            else:
                raise ValueError(f"Unexpected file scope type: {type(scope)}. Expected DataSet or All.")
        return output

    def assets(
        self,
        action: Sequence[Literal["read", "write"]],
        dataset_ids: set[int] | None = None,
        operation: str | None = None,
    ) -> dict[str, list[str]] | None:
        """Validate access to assets.
        Args:
            action (Sequence[Literal["read", "write"]]): The actions to validate access for
            dataset_ids (Set[int] | None): The dataset IDs to check access for. If None, checks access for all datasets.
            operation (str | None): The operation being performed, used for error messages.
        Returns:
            dict[str, list[str]] | None: Returns a dictionary with the key 'dataset'
                if access is limited to a dataset scope, or None if access is granted to all assets.
        Raises:
            ValueError: If the client.token.get_scope() returns an unexpected asset scope type.
            AuthorizationError: If the user does not have permission to perform the specified action on the given
                dataset.
        """
        operation = operation or self.default_operation
        asset_scopes, actions_str = self._set_up_read_write(
            action, AssetsAcl.Action.Read, AssetsAcl.Action.Write, operation, "assets"
        )
        if isinstance(asset_scopes[0], AssetsAcl.Scope.All):
            return None
        if dataset_ids is not None:
            missing = set(dataset_ids)
            for scope in asset_scopes:
                if isinstance(scope, AssetsAcl.Scope.DataSet):
                    missing = missing - set(scope.ids)
                    if not missing:
                        return None
            raise AuthorizationError(
                f"You have no permission to {actions_str} assets in dataset(s) {humanize_collection(missing)}. This is required to {operation}."
            )
        output: dict[str, list[str]] = {}
        for scope in asset_scopes:
            if isinstance(scope, AssetsAcl.Scope.DataSet):
                output["dataset"] = self.client.lookup.data_sets.external_id(scope.ids)
            else:
                raise ValueError(f"Unexpected asset scope type: {type(scope)}. Expected DataSet or All.")
        return output

    def _set_up_read_write(
        self,
        action: Sequence[Literal["read", "write"]],
        read: Capability.Action,
        write: Capability.Action,
        operation: str,
        name: str,
    ) -> tuple[list[Capability.Scope], str]:
        actions_str = humanize_collection(action, bind_word="and")
        actions = [{"read": read, "write": write}[a] for a in action]
        scopes = self.client.token.get_scope(actions)
        if scopes is None:
            raise AuthorizationError(
                f"You have no permission to {actions_str} {name}. This is required to {operation}."
            )
        return scopes, actions_str



import json
from collections.abc import Iterator
from pathlib import Path

import pytest
import responses
from cognite.client.data_classes import Asset, AssetList
from cognite.client.data_classes.data_modeling import (
    ContainerId,
    DirectRelation,
    MappedProperty,
    NodeApply,
    NodeOrEdgeData,
    Text,
    View,
    ViewId,
)

from cognite_toolkit._cdf_tk.client import ToolkitClient, ToolkitClientConfig
from cognite_toolkit._cdf_tk.commands._migrate.adapter import AssetCentricMigrationIOAdapter, MigrationCSVFileSelector
from cognite_toolkit._cdf_tk.commands._migrate.command import MigrationCommand
from cognite_toolkit._cdf_tk.commands._migrate.data_mapper import AssetCentricMapper
from cognite_toolkit._cdf_tk.commands._migrate.data_model import INSTANCE_SOURCE_VIEW_ID
from cognite_toolkit._cdf_tk.commands._migrate.default_mappings import _ASSET_ID, create_default_mappings
from cognite_toolkit._cdf_tk.storageio import AssetIO, InstanceIO


@pytest.fixture
def cognite_migration_model(
    toolkit_config: ToolkitClientConfig, rsps: responses.RequestsMock
) -> Iterator[responses.RequestsMock]:
    config = toolkit_config
    mapping_by_id = {mapping.external_id: mapping for mapping in create_default_mappings()}
    asset_mapping = mapping_by_id[_ASSET_ID]
    # Lookup of the mapping in the Migration Model
    mapping_node_response = asset_mapping.dump(context="api")
    mapping_node_response.update({"createdTime": 0, "lastUpdatedTime": 0, "version": 1})
    sources = mapping_node_response.pop("sources", [])
    if sources:
        mapping_view_id = asset_mapping.sources[0].source
        mapping_node_response["properties"] = {
            mapping_view_id.space: {
                f"{mapping_view_id.external_id}/{mapping_view_id.version}": sources[0]["properties"]
            }
        }
    rsps.post(
        config.create_api_url("models/instances/byids"),
        json={"items": [mapping_node_response]},
        status=200,
    )

    # Lookup CogniteAsset, this is not the full model, just the properties we need for the
    # migration
    default_prop_args = dict(nullable=True, immutable=False, auto_increment=False)
    default_view_args = dict(
        last_updated_time=1,
        created_time=1,
        description=None,
        name=None,
        implements=[],
        writable=True,
        used_for="node",
        is_global=True,
        filter=None,
    )
    cognite_asset = View(
        space="cdf_cdm",
        external_id="CogniteAsset",
        version="v1",
        properties={
            "name": MappedProperty(ContainerId("cdf_cdm", "CogniteDescribable"), "name", Text(), **default_prop_args),
            "description": MappedProperty(
                ContainerId("cdf_cdm", "CogniteDescribable"), "description", Text(), **default_prop_args
            ),
            "source": MappedProperty(
                ContainerId("cdf_cdm", "CogniteSourceable"), "name", DirectRelation(), **default_prop_args
            ),
            "tags": MappedProperty(
                ContainerId("cdf_cdm", "CogniteDescribable"), "tags", Text(is_list=True), **default_prop_args
            ),
        },
        **default_view_args,
    )
    rsps.post(
        config.create_api_url("models/views/byids"),
        json={"items": [cognite_asset.dump()]},
    )
    yield rsps


class TestMigrationCommand:
    @pytest.mark.usefixtures("disable_gzip", "disable_pypi_check")
    def test_migrate_assets(
        self, toolkit_config: ToolkitClientConfig, cognite_migration_model: responses.RequestsMock, tmp_path: Path
    ) -> None:
        rsps = cognite_migration_model
        config = toolkit_config
        assets = AssetList(
            [
                Asset(
                    id=1000 + i,
                    external_id=f"asset_{i}",
                    name=f"Asset {i}",
                    description=f"This is Asset {i}",
                    last_updated_time=1,
                    created_time=0,
                    parent_external_id="asset_0" if i > 0 else None,
                )
                for i in range(2)
            ]
        )
        space = "my_space"
        csv_content = "id,space,externalId,ingestionView\n" + "\n".join(
            f"{1000 + i},{space},asset_{i},{_ASSET_ID}" for i in range(len(assets))
        )

        # Asset retrieve ids
        rsps.post(
            config.create_api_url("/assets/byids"),
            json={"items": [asset.dump() for asset in assets]},
            status=200,
        )
        # Instance creation
        rsps.post(
            config.create_api_url("/models/instances"),
            json={
                "items": [
                    {
                        "instanceType": "node",
                        "space": space,
                        "externalId": f"asset_{i}",
                        "version": 1,
                        "wasModified": True,
                        "createdTime": 1,
                        "lastUpdatedTime": 1,
                    }
                    for i in range(len(assets))
                ]
            },
            status=200,
        )

        csv_file = tmp_path / "migration.csv"
        csv_file.write_text(csv_content, encoding="utf-8")

        client = ToolkitClient(config)
        command = MigrationCommand(silent=True)

        result = command.migrate(
            selected=MigrationCSVFileSelector(csv_file, resource_type="asset"),
            data=AssetCentricMigrationIOAdapter(client, AssetIO(client), InstanceIO(client)),
            mapper=AssetCentricMapper(client),
            log_dir=tmp_path / "logs",
            dry_run=False,
            verbose=False,
        )

        # Check that the assets were uploaded
        last_call = rsps.calls[-1]
        assert last_call.request.url == config.create_api_url("/models/instances")
        assert last_call.request.method == "POST"
        actual_instances = json.loads(last_call.request.body)["items"]
        expected_instance = [
            NodeApply(
                space=space,
                external_id=asset.external_id,
                sources=[
                    NodeOrEdgeData(
                        source=ViewId("cdf_cdm", "CogniteAsset", "v1"),
                        properties={
                            "name": asset.name,
                            "description": asset.description,
                        },
                    ),
                    NodeOrEdgeData(
                        source=INSTANCE_SOURCE_VIEW_ID,
                        properties={
                            "id": asset.id,
                            "resourceType": "asset",
                            "dataSetId": None,
                            "classicExternalId": asset.external_id,
                        },
                    ),
                ],
            ).dump()
            for asset in assets
        ]
        assert actual_instances == expected_instance
        actual_results = [result.get_progress(asset.id) for asset in assets]
        expected_results = [{"download": "success", "convert": "success", "upload": "success"} for _ in assets]
        assert actual_results == expected_results



from abc import ABC
from typing import TypeVar

from cognite.client.data_classes import Asset, Event, FileMetadata, Sequence, TimeSeries
from cognite.client.data_classes.capabilities import (
    Capability,
    DataModelInstancesAcl,
    DataModelsAcl,
    SpaceIDScope,
)

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.commands._base import ToolkitCommand
from cognite_toolkit._cdf_tk.constants import DMS_INSTANCE_LIMIT_MARGIN
from cognite_toolkit._cdf_tk.exceptions import (
    AuthenticationError,
    ToolkitMigrationError,
    ToolkitValueError,
)
from cognite_toolkit._cdf_tk.utils import humanize_collection

from .data_model import INSTANCE_SOURCE_VIEW_ID, MODEL_ID, RESOURCE_VIEW_MAPPING_VIEW_ID

T_AssetCentricResource = TypeVar("T_AssetCentricResource", bound=Asset | Event | FileMetadata | TimeSeries | Sequence)


class BaseMigrateCommand(ToolkitCommand, ABC):
    def source_acl(self, data_set_id: list[int]) -> Capability:
        """Return the source ACL for the given data set IDs."""
        # This method should be implemented in subclasses that needs access to a specific source ACL.
        # such as TimeSeries, Files, Assets, and so on.
        raise ValueError(
            "Bug in Toolkit: the source ACL is not defined for this migration command. "
            "Please implement the source_acl method."
        )

    def validate_access(
        self,
        client: ToolkitClient,
        instance_spaces: list[str] | None = None,
        schema_spaces: list[str] | None = None,
        data_set_ids: list[int] | None = None,
    ) -> None:
        required_capabilities: list[Capability] = []
        if instance_spaces is not None:
            required_capabilities.append(
                DataModelInstancesAcl(
                    actions=[
                        DataModelInstancesAcl.Action.Read,
                        DataModelInstancesAcl.Action.Write,
                        DataModelInstancesAcl.Action.Write_Properties,
                    ],
                    scope=SpaceIDScope(instance_spaces),
                )
            )
        if schema_spaces is not None:
            required_capabilities.append(
                DataModelsAcl(actions=[DataModelsAcl.Action.Read], scope=SpaceIDScope(schema_spaces)),
            )

        if data_set_ids is not None:
            source_acl = self.source_acl(data_set_ids)
            required_capabilities.append(source_acl)
        if missing := client.iam.verify_capabilities(required_capabilities):
            raise AuthenticationError(f"Missing required capabilities: {humanize_collection(missing)}.", missing)

    @staticmethod
    def validate_migration_model_available(client: ToolkitClient) -> None:
        models = client.data_modeling.data_models.retrieve([MODEL_ID], inline_views=False)
        if not models:
            raise ToolkitMigrationError(
                f"The migration data model {MODEL_ID!r} does not exist. "
                "Please run the `cdf migrate prepare` command to deploy the migration data model."
            )
        elif len(models) > 1:
            raise ToolkitMigrationError(
                f"Multiple migration models {MODEL_ID!r}. "
                "Please delete the duplicate models before proceeding with the migration."
            )
        model = models[0]
        missing_views = {INSTANCE_SOURCE_VIEW_ID, RESOURCE_VIEW_MAPPING_VIEW_ID} - set(model.views or [])
        if missing_views:
            raise ToolkitMigrationError(
                f"Invalid migration model. Missing views {humanize_collection(missing_views)}. "
                f"Please run the `cdf migrate prepare` command to deploy the migration data model."
            )

    def validate_available_capacity(self, client: ToolkitClient, instance_count: int) -> None:
        """Validate that the project has enough capacity to accommodate the migration."""

        stats = client.data_modeling.statistics.project()

        available_capacity = stats.instances.instances_limit - stats.instances.instances
        available_capacity_after = available_capacity - instance_count

        if available_capacity_after < DMS_INSTANCE_LIMIT_MARGIN:
            raise ToolkitValueError(
                "Cannot proceed with migration, not enough instance capacity available. Total capacity after migration"
                f" would be {available_capacity_after:,} instances, which is less than the required margin of"
                f" {DMS_INSTANCE_LIMIT_MARGIN:,} instances. Please increase the instance capacity in your CDF project"
                f" or delete some existing instances before proceeding with the migration of {instance_count:,} assets."
            )
        total_instances = stats.instances.instances + instance_count
        self.console(
            f"Project has enough capacity for migration. Total instances after migration: {total_instances:,}."
        )



from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import Generic, Literal

from cognite.client.data_classes.data_modeling import EdgeId, NodeId, ViewId
from cognite.client.utils._identifier import InstanceId

from cognite_toolkit._cdf_tk.storageio._data_classes import InstanceIdCSVList, T_ModelList
from cognite_toolkit._cdf_tk.utils.file import to_directory_compatible


@dataclass(frozen=True)
class CSVFileSelector(Generic[T_ModelList], ABC):
    """Data class for file-based data selection from CSV files."""

    datafile: Path

    @abstractmethod
    def list_cls(self) -> type[T_ModelList]:
        raise NotImplementedError()

    @cached_property
    def items(self) -> T_ModelList:
        return self.list_cls().read_csv_file(self.datafile)

    def __str__(self) -> str:
        return f"{type(self).__name__}={self.datafile.name}"


@dataclass(frozen=True)
class AssetCentricSelector:
    """Data class for asset-centric data selection."""


@dataclass(frozen=True)
class DataSetSelector(AssetCentricSelector):
    """Select data associated with a specific data set."""

    data_set_external_id: str

    def __str__(self) -> str:
        return f"DataSet={to_directory_compatible(self.data_set_external_id)}"


@dataclass(frozen=True)
class AssetSubtreeSelector(AssetCentricSelector):
    """Select data associated with an asset and its subtree."""

    hierarchy: str

    def __str__(self) -> str:
        return f"AssetSubtree={to_directory_compatible(self.hierarchy)}"


@dataclass(frozen=True)
class AssetCentricFileSelector(AssetCentricSelector):
    """Select data from a specific file."""

    datafile: Path

    def __str__(self) -> str:
        return f"File={self.datafile.name}"


@dataclass(frozen=True)
class InstanceSelector:
    @abstractmethod
    def get_schema_spaces(self) -> list[str] | None:
        raise NotImplementedError()

    @abstractmethod
    def get_instance_spaces(self) -> list[str] | None:
        raise NotImplementedError()


@dataclass(frozen=True)
class InstanceFileSelector(CSVFileSelector[InstanceIdCSVList], InstanceSelector):
    datafile: Path
    validate: bool = True

    def list_cls(self) -> type[InstanceIdCSVList]:
        return InstanceIdCSVList

    @cached_property
    def _ids_by_type(self) -> tuple[list[NodeId], list[EdgeId]]:
        node_ids: list[NodeId] = []
        edge_ids: list[EdgeId] = []
        for instance in self.items:
            if instance.instance_type == "node":
                node_ids.append(NodeId(instance.space, instance.external_id))
            else:
                edge_ids.append(EdgeId(instance.space, instance.external_id))
        return node_ids, edge_ids

    @property
    def instance_ids(self) -> list[InstanceId]:
        node_ids, edge_ids = self._ids_by_type
        return [*node_ids, *edge_ids]

    @property
    def node_ids(self) -> list[NodeId]:
        return self._ids_by_type[0]

    @property
    def edge_ids(self) -> list[EdgeId]:
        return self._ids_by_type[1]

    def get_schema_spaces(self) -> list[str] | None:
        return None

    def get_instance_spaces(self) -> list[str] | None:
        return sorted({instance.space for instance in self.items})

    def __str__(self) -> str:
        return f"file {self.datafile.as_posix()}"


@dataclass(frozen=True)
class InstanceViewSelector(InstanceSelector):
    view: ViewId
    instance_type: Literal["node", "edge"] = "node"
    instance_spaces: tuple[str, ...] | None = None

    def get_schema_spaces(self) -> list[str] | None:
        return [self.view.space]

    def get_instance_spaces(self) -> list[str] | None:
        return list(self.instance_spaces) if self.instance_spaces else None

    def __str__(self) -> str:
        return f"view {self.view!r}"


@dataclass(frozen=True)
class ChartSelector: ...


@dataclass(frozen=True)
class ChartOwnerSelector(ChartSelector):
    owner_id: str


@dataclass(frozen=True)
class AllChartSelector(ChartSelector): ...


@dataclass(frozen=True)
class ChartFileSelector(ChartSelector):
    filepath: Path



from enum import Enum
from pathlib import Path
from typing import Annotated, Any

import typer
from rich import print

from cognite_toolkit._cdf_tk.client.data_classes.raw import RawTable
from cognite_toolkit._cdf_tk.commands import DownloadCommand
from cognite_toolkit._cdf_tk.storageio import (
    AssetCentricSelector,
    AssetIO,
    AssetSubtreeSelector,
    DataSetSelector,
    RawIO,
)
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables
from cognite_toolkit._cdf_tk.utils.interactive_select import (
    AssetInteractiveSelect,
    RawTableInteractiveSelect,
)


class RawFormats(str, Enum):
    ndjson = "ndjson"
    yaml = "yaml"


class AssetCentricFormats(str, Enum):
    csv = "csv"
    parquet = "parquet"
    ndjson = "ndjson"


class CompressionFormat(str, Enum):
    gzip = "gzip"
    none = "none"


class DownloadApp(typer.Typer):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.callback(invoke_without_command=True)(self.download_main)
        self.command("raw")(self.download_raw_cmd)
        self.command("assets")(self.download_assets_cmd)

    @staticmethod
    def download_main(ctx: typer.Context) -> None:
        """Commands to download data from CDF into a temporary directory."""
        if ctx.invoked_subcommand is None:
            print("Use [bold yellow]cdf download --help[/] for more information.")
        return None

    @staticmethod
    def download_raw_cmd(
        ctx: typer.Context,
        tables: Annotated[
            list[str] | None,
            typer.Argument(
                help="List of tables to download. If not provided, an interactive selection will be made.",
            ),
        ] = None,
        database: Annotated[
            str | None,
            typer.Option(
                "--database",
                "-d",
                help="Database to download from. If not provided, the user will be prompted to select a database.",
            ),
        ] = None,
        file_format: Annotated[
            RawFormats,
            typer.Option(
                "--format",
                "-f",
                help="Format to download the raw tables in. Supported formats: ndjson, yaml",
            ),
        ] = RawFormats.ndjson,
        compression: Annotated[
            CompressionFormat,
            typer.Option(
                "--compression",
                "-z",
                help="Compression format to use when downloading the raw tables. Supported formats: gzip, none.",
            ),
        ] = CompressionFormat.gzip,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Where to download the raw tables.",
                allow_dash=True,
            ),
        ] = Path("tmp"),
        limit: Annotated[
            int,
            typer.Option(
                "--limit",
                "-l",
                help="The maximum the number of records to download from each table. Use -1 to download all records.",
            ),
        ] = 100_000,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will download RAW tables from CDF into a temporary directory."""
        cmd = DownloadCommand()

        client = EnvironmentVariables.create_from_environment().get_client()
        if tables and database:
            selectors = [RawTable(db_name=database, table_name=table) for table in tables]
        elif tables and not database:
            raise typer.BadParameter(
                "The '--database' option is required when specifying tables as arguments.",
                param_hint="--database",
            )
        elif not tables and database:
            selectors = RawTableInteractiveSelect(client, "download").select_tables(database=database)
        else:
            selectors = RawTableInteractiveSelect(client, "download").select_tables()

        cmd.run(
            lambda: cmd.download(
                selectors=selectors,
                io=RawIO(client),
                output_dir=output_dir,
                file_format=f".{file_format.value}",
                compression=compression.value,
                limit=limit if limit != -1 else None,
                verbose=verbose,
            )
        )

    @staticmethod
    def download_assets_cmd(
        ctx: typer.Context,
        data_sets: Annotated[
            list[str] | None,
            typer.Option(
                "--data-set",
                "-d",
                help="List of data sets to download assets from. If this and hierarchy are not provided, an interactive selection will be made.",
            ),
        ] = None,
        hierarchy: Annotated[
            list[str] | None,
            typer.Option(
                "--hierarchy",
                "-r",
                help="List of asset hierarchies to download assets from. If this and data sets are not provided, an interactive selection will be made.",
            ),
        ] = None,
        file_format: Annotated[
            AssetCentricFormats,
            typer.Option(
                "--format",
                "-f",
                help="Format to download the assets in.",
            ),
        ] = AssetCentricFormats.csv,
        compression: Annotated[
            CompressionFormat,
            typer.Option(
                "--compression",
                "-z",
                help="Compression format to use when downloading the assets.",
            ),
        ] = CompressionFormat.none,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Where to download the assets.",
                allow_dash=True,
            ),
        ] = Path("tmp"),
        limit: Annotated[
            int,
            typer.Option(
                "--limit",
                "-l",
                help="The maximum number of assets to download from each dataset/hierarchy. Use -1 to download all assets.",
            ),
        ] = 100_000,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will download assets from CDF into a temporary directory."""
        client = EnvironmentVariables.create_from_environment().get_client()
        is_interactive = not data_sets and not hierarchy
        if is_interactive:
            interactive = AssetInteractiveSelect(client, "download assets")
            selector_type = interactive.select_hierarchies_or_data_sets()
            if selector_type == "Data Set":
                data_sets = interactive.select_data_sets()
            else:
                hierarchy = interactive.select_hierarchies()

        selectors: list[AssetCentricSelector] = []
        if data_sets:
            selectors.extend([DataSetSelector(data_set_external_id=ds) for ds in data_sets])
        if hierarchy:
            selectors.extend([AssetSubtreeSelector(hierarchy=h) for h in hierarchy])
        cmd = DownloadCommand()
        cmd.run(
            lambda: cmd.download(
                selectors=selectors,
                io=AssetIO(client),
                output_dir=output_dir,
                file_format=f".{file_format.value}",
                compression=compression.value,
                limit=limit if limit != -1 else None,
                verbose=verbose,
            )
        )



from datetime import datetime, timezone

import pytest
from cognite.client.data_classes.data_modeling import NodeList, ViewId

from cognite_toolkit._cdf_tk.client.data_classes.canvas import Canvas, IndustrialCanvas, IndustrialCanvasApply
from cognite_toolkit._cdf_tk.client.data_classes.migration import InstanceSource
from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.commands import MigrationCanvasCommand
from cognite_toolkit._cdf_tk.commands._migrate.data_model import COGNITE_MIGRATION_MODEL
from cognite_toolkit._cdf_tk.tk_warnings import HighSeverityWarning, ToolkitWarning


@pytest.fixture(scope="session")
def asset_centric_canvas() -> tuple[IndustrialCanvas, NodeList[InstanceSource]]:
    canvas = IndustrialCanvas.load(
        {
            "annotations": [],
            "canvas": [
                {
                    "createdTime": 1751540227230,
                    "externalId": "495af88f-fe1d-403d-91b1-76ef9f80f265",
                    "instanceType": "node",
                    "lastUpdatedTime": 1751540558717,
                    "properties": {
                        "cdf_industrial_canvas": {
                            "Canvas/v7": {
                                "createdBy": "aGQ-cBXUBY6bmmxqIdkFoA",
                                "isArchived": None,
                                "isLocked": None,
                                "name": "Asset-centric1",
                                "solutionTags": None,
                                "sourceCanvasId": None,
                                "updatedAt": "2025-07-03T11:02:37.733+00:00",
                                "updatedBy": "aGQ-cBXUBY6bmmxqIdkFoA",
                                "visibility": "public",
                            }
                        }
                    },
                    "space": "IndustrialCanvasInstanceSpace",
                    "version": 14,
                }
            ],
            "containerReferences": [
                {
                    "createdTime": 1751540264906,
                    "externalId": "495af88f-fe1d-403d-91b1-76ef9f80f265_cf372b29-3012-49ff-8daf-5043404c23d7",
                    "instanceType": "node",
                    "lastUpdatedTime": 1751540264906,
                    "properties": {
                        "cdf_industrial_canvas": {
                            "ContainerReference/v2": {
                                "chartsId": None,
                                "containerReferenceType": "asset",
                                "height": 357,
                                "id": "cf372b29-3012-49ff-8daf-5043404c23d7",
                                "label": "Kelmarsh 6",
                                "maxHeight": None,
                                "maxWidth": None,
                                "resourceId": 3840956528416998,
                                "resourceSubId": None,
                                "width": 600,
                                "x": 0,
                                "y": 0,
                            }
                        }
                    },
                    "space": "IndustrialCanvasInstanceSpace",
                    "version": 1,
                },
                {
                    "createdTime": 1751540275336,
                    "externalId": "495af88f-fe1d-403d-91b1-76ef9f80f265_09d58ddf-bebb-4e4d-96db-1702da76a016",
                    "instanceType": "node",
                    "lastUpdatedTime": 1751540275336,
                    "properties": {
                        "cdf_industrial_canvas": {
                            "ContainerReference/v2": {
                                "chartsId": None,
                                "containerReferenceType": "timeseries",
                                "height": 400,
                                "id": "09d58ddf-bebb-4e4d-96db-1702da76a016",
                                "label": "Hub temperature, standard deviation (Â°C)",
                                "maxHeight": None,
                                "maxWidth": None,
                                "resourceId": 11978459264156,
                                "resourceSubId": None,
                                "width": 700,
                                "x": 700,
                                "y": 0,
                            }
                        }
                    },
                    "space": "IndustrialCanvasInstanceSpace",
                    "version": 1,
                },
                {
                    "createdTime": 1751540544349,
                    "externalId": "495af88f-fe1d-403d-91b1-76ef9f80f265_5e2bf845-103c-4c17-8549-9f20329b7f98",
                    "instanceType": "node",
                    "lastUpdatedTime": 1751540558717,
                    "properties": {
                        "cdf_industrial_canvas": {
                            "ContainerReference/v2": {
                                "chartsId": None,
                                "containerReferenceType": "event",
                                "height": 500,
                                "id": "5e2bf845-103c-4c17-8549-9f20329b7f98",
                                "label": "b18cdf8e-6568-4e2a-a267-535eb52f41bf",
                                "maxHeight": None,
                                "maxWidth": None,
                                "resourceId": 9004025980300864,
                                "resourceSubId": None,
                                "width": 600,
                                "x": -10,
                                "y": 418,
                            }
                        }
                    },
                    "space": "IndustrialCanvasInstanceSpace",
                    "version": 4,
                },
            ],
        }
    )
    mapping = NodeList[InstanceSource](
        [
            InstanceSource(
                space="MyNewInstanceSpace",
                external_id="my_asset",
                version=1,
                last_updated_time=1,
                created_time=1,
                resource_type="asset",
                id_=3840956528416998,
                preferred_consumer_view_id=ViewId("my_space", "DoctrinoAsset", "v1"),
            ),
            InstanceSource(
                space="MyNewInstanceSpace",
                external_id="my_timeseries",
                version=1,
                last_updated_time=1,
                created_time=1,
                resource_type="timeseries",
                id_=11978459264156,
                preferred_consumer_view_id=ViewId("my_space", "DoctrinoTimeSeries", "v1"),
            ),
            InstanceSource(
                space="MyNewInstanceSpace",
                external_id="my_event",
                version=1,
                last_updated_time=1,
                created_time=1,
                resource_type="event",
                id_=9004025980300864,
            ),
        ]
    )
    return canvas, mapping


class TestMigrationCanvasCommand:
    def test_migrate_canvas_happy_path(
        self, asset_centric_canvas: tuple[IndustrialCanvas, NodeList[InstanceSource]]
    ) -> None:
        command = MigrationCanvasCommand(silent=True)
        canvas, instance_sources = asset_centric_canvas

        with monkeypatch_toolkit_client() as client:
            client.iam.verify_capabilities.return_value = []
            client.data_modeling.data_models.retrieve.return_value = [COGNITE_MIGRATION_MODEL]
            client.canvas.industrial.retrieve.return_value = canvas
            client.migration.instance_source.retrieve.return_value = instance_sources

            command.migrate_canvas(client, external_ids=["my_canvas"], dry_run=False, verbose=True)

            client.canvas.industrial.update.assert_called_once()
            update = client.canvas.industrial.update.call_args[0][0]
            assert isinstance(update, IndustrialCanvasApply)
            client.canvas.industrial.create.assert_called_once()
            backup = client.canvas.industrial.create.call_args[0][0]
            assert isinstance(backup, IndustrialCanvasApply)

        assert len(update.fdm_instance_container_references) == len(canvas.container_references)
        assert len(backup.fdm_instance_container_references) == 0

    def test_migrate_canvas_missing(self) -> None:
        command = MigrationCanvasCommand(silent=True)

        with monkeypatch_toolkit_client() as client:
            client.iam.verify_capabilities.return_value = []
            client.data_modeling.data_models.retrieve.return_value = [COGNITE_MIGRATION_MODEL]
            client.canvas.industrial.retrieve.return_value = None
            command.migrate_canvas(client, external_ids=["non-existing"], dry_run=False, verbose=True)
        assert len(command.warning_list) == 1
        warning = command.warning_list[0]
        assert isinstance(warning, ToolkitWarning)
        assert "Canvas with external ID 'non-existing' not found." in str(warning)

    def test_migrate_canvas_missing_instance_source(
        self, asset_centric_canvas: tuple[IndustrialCanvas, NodeList[InstanceSource]]
    ) -> None:
        command = MigrationCanvasCommand(silent=True)
        canvas, _ = asset_centric_canvas

        with monkeypatch_toolkit_client() as client:
            client.iam.verify_capabilities.return_value = []
            client.data_modeling.data_models.retrieve.return_value = [COGNITE_MIGRATION_MODEL]
            client.canvas.industrial.retrieve.return_value = canvas
            client.migration.instance_source.retrieve.return_value = NodeList[InstanceSource]([])

            command.migrate_canvas(client, external_ids=[canvas.canvas.external_id], dry_run=False, verbose=True)

        assert len(command.warning_list) == 1
        warning = command.warning_list[0]
        assert isinstance(warning, HighSeverityWarning)
        assert "Canvas 'Asset-centric1' has references to resources that are not been migrated" in str(warning)

    def test_migrate_canvas_no_asset_centric_references(
        self, asset_centric_canvas: tuple[IndustrialCanvas, NodeList[InstanceSource]]
    ) -> None:
        command = MigrationCanvasCommand(silent=True)
        canvas = IndustrialCanvas(
            canvas=Canvas(
                "canvasSpace",
                "canvasExternalId",
                1,
                1,
                1,
                name="MyCanvas",
                created_by="me",
                updated_at=datetime(2020, 1, 1, tzinfo=timezone.utc),
                updated_by="me",
            )
        )

        with monkeypatch_toolkit_client() as client:
            client.iam.verify_capabilities.return_value = []
            client.data_modeling.data_models.retrieve.return_value = [COGNITE_MIGRATION_MODEL]
            client.canvas.industrial.retrieve.return_value = canvas

            command.migrate_canvas(client, external_ids=[canvas.canvas.external_id], dry_run=False, verbose=True)

        assert len(command.warning_list) == 1
        warning = command.warning_list[0]
        assert isinstance(warning, ToolkitWarning)
        assert "Canvas with name 'MyCanvas' does not have any asset-centric references." in str(warning)



from cognite.client.data_classes.data_modeling import NodeId, Space
from cognite.client.data_classes.data_modeling.cdm.v1 import CogniteActivityApply

from cognite_toolkit._cdf_tk.client import ToolkitClientConfig
from cognite_toolkit._cdf_tk.utils.batch_processor import HTTPIterableProcessor


class TestBatchProcessor:
    def test_batch_processor_create_nodes(
        self, toolkit_client_config: ToolkitClientConfig, toolkit_space: Space
    ) -> None:
        config = toolkit_client_config
        some_nodes = (
            CogniteActivityApply(
                space=toolkit_space.space if i < 9 else "non_existent_space",
                external_id=f"toolkit_test_batch_processor_{i}",
                name=f"Test Activity {i}",
            ).dump()
            for i in range(10)
        )
        with HTTPIterableProcessor[NodeId](
            endpoint_url=config.create_api_url("/models/instances"),
            config=config,
            as_id=lambda item: NodeId(item["space"], item["externalId"]),
            body_parameters={"autoCreateDirectRelations": True},
            method="POST",
            batch_size=1000,
            max_workers=2,
            description="Create nodes",
        ) as processor:
            result = processor.process(some_nodes)

        assert result.total_successful == 9
        assert result.total_failed == 1



from typing import Literal, cast
from urllib.parse import urljoin

from cognite.client import ClientConfig, CogniteClient
from cognite.client.credentials import CredentialProvider

from .api.canvas import CanvasAPI
from .api.charts import ChartsAPI
from .api.dml import DMLAPI
from .api.extended_data_modeling import ExtendedDataModelingAPI
from .api.extended_files import ExtendedFileMetadataAPI
from .api.extended_functions import ExtendedFunctionsAPI
from .api.extended_raw import ExtendedRawAPI
from .api.extended_timeseries import ExtendedTimeSeriesAPI
from .api.lookup import LookUpGroup
from .api.migration import MigrationAPI
from .api.robotics import RoboticsAPI
from .api.search import SearchAPI
from .api.token import TokenAPI
from .api.verify import VerifyAPI


class ToolkitClientConfig(ClientConfig):
    def __init__(
        self,
        client_name: str,
        project: str,
        credentials: CredentialProvider,
        api_subversion: str | None = None,
        base_url: str | None = None,
        max_workers: int | None = None,
        is_strict_validation: bool = True,
        headers: dict[str, str] | None = None,
        timeout: int | None = None,
        file_transfer_timeout: int | None = None,
        debug: bool = False,
    ) -> None:
        super().__init__(
            client_name=client_name,
            project=project,
            credentials=credentials,
            api_subversion=api_subversion,
            base_url=base_url,
            max_workers=max_workers,
            headers=headers,
            timeout=timeout,
            file_transfer_timeout=file_transfer_timeout,
            debug=debug,
        )
        self.is_strict_validation = is_strict_validation

    @property
    def cloud_provider(self) -> Literal["azure", "aws", "gcp", "unknown"]:
        cdf_cluster = self.cdf_cluster
        if cdf_cluster is None:
            return "unknown"
        elif cdf_cluster.startswith("az-") or cdf_cluster in {"azure-dev", "bluefield", "westeurope-1"}:
            return "azure"
        elif cdf_cluster.startswith("aws-") or cdf_cluster in {"orangefield"}:
            return "aws"
        elif cdf_cluster.startswith("gc-") or cdf_cluster in {
            "greenfield",
            "asia-northeast1-1",
            "cognitedata-development",
            "cognitedata-production",
        }:
            return "gcp"
        else:
            return "unknown"

    @property
    def is_private_link(self) -> bool:
        if "cognitedata.com" not in self.base_url:
            return False
        subdomain = self.base_url.split("cognitedata.com", maxsplit=1)[0]
        return "plink" in subdomain

    def create_api_url(self, endpoint: str) -> str:
        """Create a full API URL for the given endpoint.

        Args:
            endpoint (str): The API endpoint to append to the base URL.

        Returns:
            str: The full API URL.

        Examples:
            >>> config = ToolkitClientConfig(cluster="bluefield", project="my_project", ...)
            >>> config.create_api_url("/models/instances")
            "https://bluefield.cognitedata.com/api/v1/my_project/models/instances"
        """
        if not endpoint.startswith("/"):
            endpoint = f"/{endpoint}"
        return f"{self.base_url}/api/v1/projects/{self.project}{endpoint}"

    def create_app_url(self, endpoint: str) -> str:
        """Create a full App URL for the given endpoint.

        Args:
            endpoint (str): The App endpoint to append to the base URL.

        Returns:
            str: The full App URL.

        Examples:
            >>> config = ToolkitClientConfig(cluster="bluefield", project="my_project", ...)
            >>> config.create_app_url("/some/app/endpoint")
            "https://bluefield.cognitedata.com/apps/v1/projects/my_project/some/app/endpoint"
        """
        if not endpoint:
            raise ValueError("Endpoint must be a non-empty string")
        if not endpoint.startswith("/"):
            endpoint = f"/{endpoint}"
        base_path = f"/apps/v1/projects/{self.project}{endpoint}"
        return urljoin(self.base_url, base_path)


class ToolkitClient(CogniteClient):
    def __init__(self, config: ToolkitClientConfig | None = None, enable_set_pending_ids: bool = False) -> None:
        super().__init__(config=config)
        self.search = SearchAPI(self._config, self._API_VERSION, self)
        self.robotics = RoboticsAPI(self._config, self._API_VERSION, self)
        self.dml = DMLAPI(self._config, self._API_VERSION, self)
        self.verify = VerifyAPI(self._config, self._API_VERSION, self)
        self.lookup = LookUpGroup(self._config, self._API_VERSION, self)
        self.functions: ExtendedFunctionsAPI = ExtendedFunctionsAPI(self._config, self._API_VERSION, self)
        self.data_modeling: ExtendedDataModelingAPI = ExtendedDataModelingAPI(self._config, self._API_VERSION, self)
        if enable_set_pending_ids:
            self.time_series: ExtendedTimeSeriesAPI = ExtendedTimeSeriesAPI(self._config, self._API_VERSION, self)
            self.files: ExtendedFileMetadataAPI = ExtendedFileMetadataAPI(self._config, self._API_VERSION, self)
        self.raw: ExtendedRawAPI = ExtendedRawAPI(self._config, self._API_VERSION, self)
        self.canvas = CanvasAPI(self.data_modeling.instances)
        self.migration = MigrationAPI(self.data_modeling.instances)
        self.token = TokenAPI(self)
        self.charts = ChartsAPI(self._config, self._API_VERSION, self)

    @property
    def config(self) -> ToolkitClientConfig:
        """Returns a config object containing the configuration for the current client.

        Returns:
            ToolkitClientConfig: The configuration object.
        """
        return cast(ToolkitClientConfig, self._config)
